import starsim as ss
import numpy as np
import pandas as pd
import datetime as dt
import scipy.stats as sps
import sciris as sc
import seaborn as sns
import matplotlib.pyplot as plt

__all__ = ['linear_interp', 'linear_accum', 'step_containing'] # Conformers
__all__ += ['CalibComponent'] # Calib component base class
__all__ += ['BetaBinomial', 'Binomial', 'DirichletMultinomial', 'GammaPoisson', 'Normal'] # Specific calib components

def linear_interp(expected, actual):
    """
    Simply interpolate, use for prevalent (stock) data like prevalence

    Args:
        expected (pd.DataFrame): The expected data from field observation, must have 't' in the index and columns corresponding to specific needs of the selected component.
        actual (pd.DataFrame): The actual data from the simulation, must have 't' in the index and columns corresponding to specific needs of the selected component.
    """
    t = expected.index
    conformed = pd.DataFrame(index=expected.index)
    for k in actual:
        conformed[k] = np.interp(x=t, xp=actual.index, fp=actual[k])

    return conformed

def step_containing(expected, actual):
    """
    Find the step containing the the timepoint.  Use for prevalent data like
    prevalence where you want to match a specific time point rather than
    interpolate.

    Args:
        expected (pd.DataFrame): The expected data from field observation, must have 't' in the index and columns corresponding to specific needs of the selected component.
        actual (pd.DataFrame): The actual data from the simulation, must have 't' in the index and columns corresponding to specific needs of the selected component.
    """
    t = expected.index
    inds = np.searchsorted(actual.index, t, side='left')
    conformed = pd.DataFrame(index=expected.index)
    for k in actual:
        conformed[k] = actual[k].values[inds] # .values because indices differ

    return conformed

def linear_accum(expected, actual):
    """
    Interpolate in the cumulative sum, then difference. Use for incident data
    (flows) like incidence or new_deaths. The accumulation is done between 't'
    and 't1', both of which must be present in the index of expected and actual
    dataframes.

    Args:
        expected (pd.DataFrame): The expected data from field observation, must have 't' and 't1' in the index and columns corresponding to specific needs of the selected component.
        actual (pd.DataFrame): The actual data from the simulation, must have 't' and 't1' in the index and columns corresponding to specific needs of the selected component.
    """
    t0 = expected.index.get_level_values('t')
    t1 = expected.index.get_level_values('t1')
    sim_t = actual.index

    fp = actual.cumsum()
    ret = {}
    for c in fp.columns:
        vals = fp[c].values.flatten()
        v0 = np.interp(x=t0, xp=sim_t, fp=vals) # Cum value at t0
        v1 = np.interp(x=t1, xp=sim_t, fp=vals) # Cum value at t1

        # Difference between end of step t1 and end of step t
        ret[c] = v1 - v0

    df = pd.DataFrame(ret, index=expected.index)
    return df


class CalibComponent(sc.prettyobj):
    """
    A class to compare a single channel of observed data with output from a
    simulation. The Calibration class can use several CalibComponent objects to
    form an overall understanding of how will a given simulation reflects
    observed data.

    Args:
        name (str) : the name of this component. Importantly, if
            extract_fn is None, the code will attempt to use the name, like
            "hiv.prevalence" to automatically extract data from the simulation.
        expected (df) : pandas DataFrame containing calibration data. The index should be the time 't' in either floating point years or datetime.
        extract_fn (callable) : a function to extract predicted/actual data in the same  
            format and with the same columns as `expected`.
        conform (str | callable): specify how to handle timepoints that don't
            align exactly between the expected and the actual/predicted/simulated
            data so they "conform" to a common time grid. Whether the data represents
            a 'prevalent' or an 'incident' quantity impacts how this alignment is performed.

            If 'prevalent', it means data in expected & actual dataframes represent
            the current state of the system, stocks like the number of currently infected
            individuals. In this case, the data in 'simulated' or 'actual' will be interpolated
            to match the timepoints in 'expected', allowing for pointwise comparisons
            between the expected and actual data. 

            If 'incident', it means data in expected & actual dataframes represent the accumulation
            of system states over a period of time, flows like the incidence of new infections. In
            this case, teh data in 'simulated' or 'actual' will be interpolated at the start ('t')
            and the end ('t1') of the period of interest in 'expected'. The difference between these
            two interpolated values will be used for comparison.

            Finally, 'step_containing' is a special case for prevalent data where the actual data is
            interpolated using a "zero order hold" method. This means that the value of the actual (simulated) data
            is matched to the timepoint in the expected data that contains the timepoint of the actual data.

        weight (float): The weight applied to the log likelihood of this component. The total log likelihood is the sum of the log likelihoods of all components, each multiplied by its weight.
        include_fn (callable): A function accepting a single simulation and returning a boolean to determine if the simulation should be included in the current component. If None, all simulations are included.
        n_boot (int): Experimental! Bootstrap sum sim results over seeds before comparing against expected results. Not appropriate for all component types.
        combine_reps (str): How to combine multiple repetitions of the same pars. Options are None, 'mean', 'sum', or other such operation. Default is None, which evaluates replicates independently instead of first combining before likelihood evaluation.
        kwargs: Additional arguments to pass to the likelihood function
    """
    def __init__(self, name, expected, extract_fn, conform, weight=1, include_fn=None, n_boot=None, combine_reps=None):
        self.name = name
        self.expected = expected
        self.extract_fn = extract_fn
        self.weight = weight
        self.include_fn = include_fn
        self.n_boot = n_boot

        self.combine_reps = combine_reps
        self.combine_kwargs = dict()
        if isinstance(self.combine_reps, str) and hasattr(pd.core.groupby.DataFrameGroupBy, self.combine_reps):
            # Most of these methods take numeric_only, which can help with stability
            self.combine_kwargs = dict(numeric_only=True)

        self.avail_conforms = {
            'none': None, # passthrough
            'incident':  linear_accum,  # or self.linear_accum if left as staticmethod  
            'prevalent': linear_interp,
            'step_containing': step_containing,
        }
        self.conform = self._validate_conform(conform)  
        return 

    def _validate_conform(self, conform):  
        ''' Validate the conform argument '''
        if not isinstance(conform, str) and not callable(conform):  
            raise Exception(f"The conform argument must be a string or a callable function, not {type(conform)}.")  
        elif isinstance(conform, str):  
            conform_ = self.avail_conforms.get(conform.lower(), 'NOT FOUND')
            if conform_ == 'NOT FOUND':
                avail = self.avail_conforms.keys()  
                raise ValueError(f"The conform argument must be one of {avail}, not {conform}.")  
        else:  
            conform_ = conform  
        return conform_

    def _combine_reps_nll(self, expected, actual, **kwargs):
        if self.combine_reps is None:
            nll = self.compute_nll(expected, actual, **kwargs) # Negative log likelihood
        else:
            timecols = [c for c in self.actual.columns if isinstance(self.actual[c].iloc[0], dt.datetime)] # Not robust to data types
            actual_combined = actual.groupby(timecols).aggregate(func=self.combine_reps, **self.combine_kwargs)
            actual_combined['rand_seed'] = 0 # Fake the seed
            actual_combined = actual_combined.reset_index().set_index('rand_seed') # Make it look like self.actual
            nll = self.compute_nll(expected, actual_combined, **kwargs)
        return nll

    def eval(self, sim, **kwargs):
        """ Compute and return the negative log likelihood """
        actuals = []
        if isinstance(sim, ss.MultiSim):
            for s in sim.sims:
                if self.include_fn is not None and not self.include_fn(s):
                    continue # Skip this simulation
                actual = self.extract_fn(s)
                if self.conform is not None:
                    actual = self.conform(self.expected, actual) # Conform
                actual['rand_seed'] = s.pars.rand_seed
                actuals.append(actual)

            if len(actuals) == 0: # No sims met the include criteria
                self.actual = None
                return np.inf
        else:
            # Warn if the user has an include_fn for a single sim
            if self.include_fn is not None and not self.include_fn(sim):
                sc.printv(f'Warning: include_fn was specified for a single simulation, but will be ignored', level=1)
            actual = self.extract_fn(sim) # Extract
            if self.conform is not None:
                actual = self.conform(self.expected, actual) # Conform
            actual['rand_seed'] = sim.pars.rand_seed
            actuals = [actual]

        self.actual = pd.concat(actuals).reset_index().set_index('rand_seed')
        seeds = self.actual.index.unique()

        if self.n_boot is None or self.n_boot == 1 or len(seeds) == 1:
            self.nll = self._combine_reps_nll(self.expected, self.actual, **kwargs)
            if self.weight == 0: # Resolve possible 0 * inf
                return 0
            wnll = self.weight * np.mean(self.nll)
            if np.isnan(wnll):
                return np.inf # Convert nan to inf
            return wnll

        # Bootstrapped aggregation
        boot_size = len(seeds)
        nlls = np.zeros(self.n_boot)
        for bi in range(self.n_boot):
            use_seeds = np.random.choice(seeds, boot_size, replace=True)
            actual = self.actual.loc[use_seeds]
            nll = self._combine_reps_nll(self.expected, actual, **kwargs)
            nlls[bi] = np.mean(nll) # Mean across reps
        self.nll = np.mean(nlls) # Mean across bootstraps

        if self.weight == 0:
            return 0 # Resolve possible 0 * inf

        wnll = self.weight * self.nll
        if np.isnan(wnll):
            return np.inf # Convert nan to inf
        return wnll

    def __call__(self, sim, **kwargs):
        return self.eval(sim, **kwargs)

    def __repr__(self):
        return f'Calibration component with name {self.name}'

    def plot(self, actual=None, bootstrap=False, **kwargs):
        actual = self.actual if actual is None else actual
        if actual is None:
            return None # Nothing to do

        if 'calibrated' not in actual.columns:
            actual['calibrated'] = 'Calibration'

        g = sns.FacetGrid(data=actual.reset_index(), col='t', row='calibrated', sharex='col', sharey=False, margin_titles=True, height=3, aspect=1.5)

        if bootstrap:
            g.map_dataframe(self.plot_facet_bootstrap)
        else:
            g.map_dataframe(self.plot_facet)
        g.set_titles(row_template='{row_name}')
        for (row_val, col_val), ax in g.axes_dict.items():
            if row_val == g.row_names[0] and isinstance(col_val, dt.datetime):
                ax.set_title(col_val.strftime('%Y-%m-%d'))

        g.fig.subplots_adjust(top=0.8)
        g.fig.suptitle(self.name)
        return g.fig

class BetaBinomial(CalibComponent):
    def compute_nll(self, expected, actual, **kwargs):
        """
        For the beta-binomial negative log-likelihood, we begin with a Beta(1,1) prior
        and subsequently observe actual['x'] successes (positives) in actual['n'] trials (total observations).
        The result is a Beta(actual['x']+1, actual['n']-actual['x']+1) posterior.
        We then compare this to the real data, which has expected['x'] successes (positives) in expected['n'] trials (total observations).
        To do so, we use a beta-binomial likelihood:
        p(x|n, a, b) = (n choose x) B(x+a, n-x+b) / B(a, b)
        where
          x=expected['x']
          n=expected['n']
          a=actual['x']+1
          b=actual['n']-actual['x']+1 
        and B is the beta function, B(x, y) = Gamma(x)Gamma(y)/Gamma(x+y)

        We return the log of p(x|n, a, b)

        kwargs will contain any eval_kwargs that were specified when instantiating the Calibration
        """

        logLs = []

        combined = pd.merge(expected.reset_index(), actual.reset_index(), on=['t'], suffixes=('_e', '_a'))
        for idx, rep in combined.iterrows():
            e_n, e_x = rep['n_e'], rep['x_e']
            a_n, a_x = rep['n_a'], rep['x_a']

            logL = sps.betabinom.logpmf(k=e_x, n=e_n, a=a_x+1, b=a_n-a_x+1)
            logLs.append(logL)

        nlls = -np.array(logLs)
        return nlls

    def plot_facet(self, data, color, **kwargs):
        t = data.iloc[0]['t']
        expected = self.expected.loc[t]
        e_n, e_x = expected['n'], expected['x']
        kk = np.arange(0, int(2*e_x))
        for idx, row in data.iterrows():
            alpha = row['x'] + 1
            beta = row['n'] - row['x'] + 1
            q = sps.betabinom(n=e_n, a=alpha, b=beta)
            yy = q.pmf(kk)
            plt.step(kk, yy, label=f"{row['rand_seed']}")
            yy = q.pmf(e_x)
            plt.plot(e_x, yy, 'x', ms=10, color='k')
        plt.axvline(e_x, color='k', linestyle='--')
        return

    def plot_facet_bootstrap(self, data, color, **kwargs):
        t = data.iloc[0]['t']
        expected = self.expected.loc[t]
        e_n, e_x = expected['n'], expected['x']

        n_boot = kwargs.get('n_boot', 1000)
        seeds = data['rand_seed'].unique()
        boot_size = len(seeds)
        means = np.zeros(n_boot)
        for bi in np.arange(n_boot):
            use_seeds = np.random.choice(seeds, boot_size, replace=True)
            if self.combine_reps is None:
                actual = data.set_index('rand_seed').loc[use_seeds]
            else:
                actual = data.set_index('rand_seed').loc[use_seeds].groupby('t').aggregate(func=self.combine_reps, **self.combine_kwargs)

            for row in actual.iterrows():
                alpha = row['x'] + 1
                beta = row['n'] - row['x'] + 1
                q = sps.betabinom(n=e_n, a=alpha, b=beta)
                means[bi] = q.mean()

        ax = sns.kdeplot(means)
        sns.rugplot(means, ax=ax)
        ax.axvline(e_x, color='k', linestyle='--')
        return

class Binomial(CalibComponent):

    @staticmethod
    def get_p(df, x_col='x', n_col='n'):
        if 'p' in df:
            p = df['p'].values
        else:
            p = df[x_col] / df[n_col] # Switched to MLE x/n from previous "Bayesian" (Laplace +1, Jeffreys) before: (df[x_col]+1) / (df[n_col]+2)
        return p

    def compute_nll(self, expected, actual, **kwargs):
        """
        Binomial log likelihood component.
        We return the log of p(x|n, p)

        kwargs will contain any eval_kwargs that were specified when instantiating the Calibration
        """

        logLs = []

        combined = pd.merge(expected.reset_index(), actual.reset_index(), on=['t'], suffixes=('_e', '_a'))
        for idx, rep in combined.iterrows():
            if 'p' in rep:
                # p specified, no collision
                e_n, e_x = rep['n'], rep['x']
                p = self.get_p(rep)
            else:
                assert 'n_e' in rep and 'x_e' in rep, 'Expected columns n_e and x_e not found'
                # Collision in merge, get _e and _a values
                e_n, e_x = rep['n_e'], rep['x_e']
                if rep['n_a'] == 0:
                    return np.inf
                p = self.get_p(rep, 'x_a', 'n_a')

            logL = sps.binom.logpmf(k=e_x, n=e_n, p=p)
            logLs.append(logL)

        nlls = -np.array(logLs)
        return nlls

    def plot_facet(self, data, color, **kwargs):
        t = data.iloc[0]['t']
        expected = self.expected.loc[t]
        e_n, e_x = expected['n'], expected['x']
        kk = np.arange(0, int(2*e_x))
        for idx, row in data.iterrows():
            p = self.get_p(row)
            q = sps.binom(n=e_n, p=p)
            yy = q.pmf(kk)
            plt.step(kk, yy, label=f"{row['rand_seed']}")
            yy = q.pmf(e_x)
            plt.plot(e_x, yy, 'x', ms=10, color='k')
        plt.axvline(e_x, color='k', linestyle='--')
        return

    def plot_facet_bootstrap(self, data, color, **kwargs):
        t = data.iloc[0]['t']
        expected = self.expected.loc[t]
        e_n, e_x = expected['n'], expected['x']

        n_boot = kwargs.get('n_boot', 1000)
        seeds = data['rand_seed'].unique()
        boot_size = len(seeds)
        means = np.zeros(n_boot)
        for bi in np.arange(n_boot):
            use_seeds = np.random.choice(seeds, boot_size, replace=True)
            if self.combine_reps is None:
                actual = data.set_index('rand_seed').loc[use_seeds]
            else:
                actual = data.set_index('rand_seed').loc[use_seeds].groupby('t').aggregate(func=self.combine_reps, **self.combine_kwargs)

            for idx, row in actual.iterrows():
                p = self.get_p(row)
                q = sps.binom(n=e_n, p=p)
                means[bi] = q.mean()

        ax = sns.kdeplot(means)
        sns.rugplot(means, ax=ax)
        ax.axvline(e_x, color='k', linestyle='--')
        return

class DirichletMultinomial(CalibComponent):
    def compute_nll(self, expected, actual, **kwargs):
        """
        The Dirichlet-multinomial negative log-likelihood is the
        multi-dimensional analog of the beta-binomial likelihood. We begin with
        a Dirichlet(1,1,...,1) prior and subsequently observe:
            actual['x1'], actual['x2'], ..., actual['xk']
        successes (positives). The result is a
            Dirichlet(alpha=actual[x_vars]+1)
        We then compare this to the real data, which has outcomes:
            expected['x1'], expected['x2'], ..., expected['xk']

        The count variables are any keys in the expected dataframe that start with 'x'.

        kwargs will contain any eval_kwargs that were specified when instantiating the Calibration
        """

        logLs = []
        x_vars = [xkey for xkey in expected.columns if xkey.startswith('x')]
        for t, rep_t in actual.groupby('t'):
            for idx, rep in rep_t.iterrows():
                e_x = expected.loc[t, x_vars].values.flatten()
                n = e_x.sum()
                a_x = rep[x_vars].astype('float')
                logL = sps.dirichlet_multinomial.logpmf(x=e_x, n=n, alpha=a_x+1)
                logLs.append(logL)

        nlls = -np.array(logLs)
        return nlls

    def plot(self, actual=None, **kwargs):
        if actual is None:
            actual = self.actual

        if 'calibrated' not in actual.columns:
            actual['calibrated'] = 'Calibration'

        x_vars = [xkey for xkey in self.expected.columns if xkey.startswith('x')]
        actual = actual \
            .reset_index() \
            [['t', 'calibrated', 'rand_seed']+x_vars] \
            .melt(id_vars=['t', 'calibrated', 'rand_seed'], var_name='var', value_name='x')
        g = sns.FacetGrid(data=actual, col='t', row='var', hue='calibrated', sharex=False, height=2, aspect=1.7, **kwargs)
        g.map_dataframe(self.plot_facet, full_actual=actual)
        g.set_titles(row_template='{row_name}')
        for (row_val, col_val), ax in g.axes_dict.items():
            if row_val == g.row_names[0] and isinstance(col_val, dt.datetime):
                ax.set_title(col_val.strftime('%Y-%m-%d'))

        g.fig.subplots_adjust(top=0.9)
        g.fig.suptitle(self.name)
        return g.fig

    def plot_facet(self, data, color, **kwargs):
        # It's challenging to plot the Dirichlet-multinomial likelihood, so we use Beta binomial as a stand-in
        t = data.iloc[0]['t']
        var = data.iloc[0]['var']
        cal = data.iloc[0]['calibrated']

        expected = self.expected.loc[[t]]
        actual = kwargs.get('full_actual', self.actual)
        actual = actual[(actual['t'] == t) & (actual['calibrated'] == cal)]

        x_vars = [xkey for xkey in expected.columns if xkey.startswith('x')]

        e_x = expected[var].values.flatten()[0]
        e_n = expected[x_vars].sum(axis=1)

        kk = np.arange(0, int(2*e_x))
        for seed, row in actual.groupby('rand_seed'):
            a = row.set_index('var')['x']
            a_x = a[var]
            a_n = a[x_vars].sum()
            alpha = a_x + 1
            beta = a_n - a_x + 1
            q = sps.betabinom(n=e_n, a=alpha, b=beta)
            yy = q.pmf(kk)
            plt.step(kk, yy, color=color, label=f"{seed}")
            yy = q.pmf(e_x)
            darker = [0.8*c for c in color]
            plt.plot(e_x, yy, 'x', ms=10, color=darker)
        plt.axvline(e_x, color='k', linestyle='--')
        return

    def plot_facet_bootstrap(self, data, color, **kwargs):
        return

class GammaPoisson(CalibComponent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        assert self.expected['n'].dtype == int, 'The expected must have an integer column named "n" for the total number of person-years'
        assert self.expected['x'].dtype == int, 'The expected must have an integer column named "x" for the total number of events'
        return

    def compute_nll(self, expected, actual, **kwargs):
        """
        The gamma-poisson likelihood is a Poisson likelihood with a
        gamma-distributed rate parameter. Through a parameter transformation, we
        end up calling a negative binomial, which is functionally equivalent.
        
        For the gamma-poisson negative log-likelihood, we begin with a
        gamma(1,1) and subsequently observe actual['x'] events (positives) in
        actual['n'] trials (total person years). The result is a
        gamma(alpha=1+actual['x'], beta=1+actual['n']) posterior.

        To evaluate the gamma-poisson likelihood as a negative binomial, we use
        the following: n = alpha, p = beta / (beta + 1)

        However, the gamma is estimating a rate, and the observation is a count over a some period of time, say T person-years. To account for T other than 1, we scale beta by 1/T to arrive at beta' = beta / T. When passing into the negative binomial, alpha remains the same, but p = beta' / (beta' + 1) or equivalently p = beta / (beta + T).

        kwargs will contain any eval_kwargs that were specified when instantiating the Calibration
        """
        logLs = []

        combined = pd.merge(expected.reset_index(), actual.reset_index(), on=['t', 't1'], suffixes=('_e', '_a'))
        for idx, rep in combined.iterrows():
            e_n, e_x = rep['n_e'], rep['x_e']
            a_n, a_x = rep['n_a'], rep['x_a']
            T = e_n
            beta = 1 + a_n
            logL = sps.nbinom.logpmf(k=e_x, n=1+a_x, p=beta/(beta+T))
            logL = np.nan_to_num(logL, nan=-np.inf)
            logLs.append(logL)

        nlls = -np.array(logLs)
        return nlls

    def plot_facet(self, data, color, **kwargs):
        t = data.iloc[0]['t']
        expected = self.expected.loc[[t]]
        e_n, e_x = expected['n'].values.flatten()[0], expected['x'].values.flatten()[0]
        kk = np.arange(0, int(2*e_x))
        nll = 0
        for idx, row in data.iterrows():
            a_n, a_x = row['n'], row['x']
            beta = (a_n+1)
            T = e_n
            q = sps.nbinom(n=1+a_x, p=beta/(beta+T))

            yy = q.pmf(kk)
            plt.step(kk, yy, label=f"{row['rand_seed']}")
            yy = q.pmf(e_x)
            nll += -q.logpmf(e_x)
            plt.plot(e_x, yy, 'x', ms=10, color='k')
        plt.axvline(e_x, color='k', linestyle='--')
        return

    def plot_facet_bootstrap(self, data, color, **kwargs):
        t = data.iloc[0]['t']
        expected = self.expected.loc[[t]]
        e_n, e_x = expected['n'].values.flatten()[0], expected['x'].values.flatten()[0]

        n_boot = kwargs.get('n_boot', 1000)
        seeds = data['rand_seed'].unique()
        boot_size = len(seeds)
        means = np.zeros(n_boot)
        for bi in np.arange(n_boot):
            use_seeds = np.random.choice(seeds, boot_size, replace=True)

            if self.combine_reps is None:
                actual = data.set_index('rand_seed').loc[use_seeds]
            else:
                actual = data.set_index('rand_seed').loc[use_seeds].groupby(['t', 't1']).aggregate(func=self.combine_reps, **self.combine_kwargs)

            for idx, row in actual.iterrows():
                a_n, a_x = row['n'], row['x']
                beta = (a_n+1)
                T = e_n
                q = sps.nbinom(n=1+a_x, p=beta/(beta+T))
                means[bi] = q.mean()

        ax = sns.kdeplot(means)
        sns.rugplot(means, ax=ax)
        plt.axvline(e_x, color='k', linestyle='--')
        return


class Normal(CalibComponent):
    def __init__(self, name, expected, extract_fn, conform, weight=1, sigma2=None, **kwargs):
        super().__init__(name, expected, extract_fn, conform, weight, **kwargs)
        self.sigma2 = sigma2
        return

    def compute_nll(self, expected, actual, **kwargs):
        """
        Normal log-likelihood component.
        
        Note that minimizing the negative log likelihood of a Gaussian likelihood is
        equivalent to minimizing the Euclidean distance between the expected and
        actual values.

        User-provided variance can be passed as a keyword argument named sigma2 when creating this component.

        Args:
            expected (pd.DataFrame): dataframe with column "x", the quantity or metric of interest, from the reference dataset.
            predicted (pd.DataFrame): dataframe with column "x", the quantity or metric of interest, from simulated dataset.
            kwargs (dict): contains any eval_kwargs that were specified when instantiating the Calibration

        Returns:
            nll (float): negative Euclidean distance between expected and predicted values.
        """

        logLs = []
        sigma2 = self.sigma2
        compute_var = sigma2 is None

        combined = pd.merge(expected.reset_index(), actual.reset_index(), on=['t'], suffixes=('_e', '_a'))
        for idx, rep in combined.iterrows():
            e_x = rep['x_e']
            a_x = rep['x_a']

            # TEMP TODO calculate rate if 'n' supplied
            if 'n' in rep:
                a_x = rep['x_a'] / rep['n']

            if compute_var:
                sigma2 = self.compute_var(expected['x'], a_x)

            logL = sps.norm.logpdf(x=e_x, loc=a_x, scale=np.sqrt(sigma2))
            logLs.append(logL)

        nlls = -np.array(logLs)
        return nlls

    def compute_var(self, expected_x, actual_x):
        """
        Compute the maximum-likelihood variance of the residuals between expected and actual values.
        """
        diffs = expected_x - actual_x
        SSE = np.sum(diffs**2)
        N = len(expected_x) if sc.isiterable(expected_x) else 1
        sigma2 = SSE/N
        return sigma2

    def plot_facet(self, data, color, **kwargs):
        t = data.iloc[0]['t']
        expected = self.expected.loc[[t]]
        e_x = expected['x'].values.flatten()[0]
        nll = 0
        for idx, row in data.iterrows():
            a_x = row['x']

            # TEMP TODO calculate rate if 'n' supplied
            if 'n' in row:
                a_x = row['x'] / row['n'] # row[['x']].values[0] / row[['n']].values[0]

            sigma2 = self.sigma2 if self.sigma2 is not None else self.compute_var(e_x, a_x)
            if isinstance(sigma2, (list, np.ndarray)):
                assert len(sigma2) == len(self.expected), 'Length of sigma2 must match the number of timepoints'
                # User provided a vector of variances
                ti = self.expected.index.get_loc(t)
                sigma2 = sigma2[ti]

            sigma = np.sqrt(sigma2)
            kk = np.linspace(a_x - 1.96*sigma, a_x + 1.96*sigma, 1000)
            q = sps.norm(loc=a_x, scale=sigma)

            yy = q.pdf(kk)
            plt.step(kk, yy, label=f"{row['rand_seed']}")
            yy = q.pdf(e_x)
            nll += -q.logpdf(e_x)
            plt.plot(e_x, yy, 'x', ms=10, color='k')
        plt.axvline(e_x, color='k', linestyle='--')
        return

    def plot_facet_bootstrap(self, data, color, **kwargs):
        t = data.iloc[0]['t']
        expected = self.expected.loc[[t]] # Gracefully handle Series and DataFrame, if 't1' in index
        e_x = expected['x'].values.flatten()[0] # Due to possible presence of 't1' in the index

        n_boot = kwargs.get('n_boot', 1000)
        seeds = data['rand_seed'].unique()
        boot_size = len(seeds)
        means = np.zeros(n_boot)
        for bi in np.arange(n_boot):
            use_seeds = np.random.choice(seeds, boot_size, replace=True)

            if self.combine_reps is None:
                actual = data.set_index('rand_seed').loc[use_seeds]
            else:
                actual = data.set_index('rand_seed').loc[use_seeds].groupby('t').aggregate(func=self.combine_reps, **self.combine_kwargs)

            for idx, row in actual.iterrows():
                a_x = row['x']

                # TEMP TODO calculate rate if 'n' supplied
                if 'n' in row:
                    a_x = row['x'] / row['n'] #row['x'].values / row['n'].values

                # No need to form the sps.norm involving sigma2 because the mean will be a_x
                means[bi] = a_x

        ax = sns.kdeplot(means)
        sns.rugplot(means, ax=ax)
        plt.axvline(e_x, color='k', linestyle='--')
        return



"""
Network connections proof of concept
"""

# %% Imports and settings
import starsim as ss
import matplotlib.pyplot as plt


ppl = ss.People(10000)

# This example uses a connector between the MF and MSM networks.
# The connector overrides participation rates so that a given
# percentage of males (prop_bi) in the MSM network are also in the
# MF network.
# This particular connector doesn't do anything with concurrency,
# i.e. a male who participates in both networks and is currently in
# a relationship in the MSM network is still considered available for
# partnerships in the MF network.
mf_pars = {'part_rates': 0.85}
msm_pars = {'part_rates': 0.1}
ppl.networks = ss.Networks(
    ss.MSMNet(pars=msm_pars), ss.MFNet(pars=mf_pars), ss.MaternalNet(),
    connectors=ss.MF_MSM(pars={'prop_bi': 0.4})
)

hiv = ss.HIV()
hiv.pars['beta'] = {'mf': [0.0008, 0.0004], 'msm': [0.004, 0.004], 'maternal': [0.2, 0]}
 
sim = ss.Sim(people=ppl, demographics=ss.Pregnancy(), diseases=[hiv, ss.Gonorrhea()])
sim.initialize()
sim.run()

plt.figure()
plt.plot(sim.yearvec, sim.results.hiv.n_infected)
plt.xlabel('Year')
plt.ylabel('Count')
plt.title('HIV number of infections')
plt.show()


"""
Test births and deaths
"""

# %% Imports and settings
import starsim as ss
import matplotlib.pyplot as plt
import pandas as pd


ppl = ss.People(10000)#, age_data=pd.read_csv('../test_data/nigeria_age.csv'))

# Parameters
simple_birth = {'birth_rates': 20}
simple_death = {'death_rates': 0.015}
simple_fertility = {'fertility_rates': 25}

realistic_birth = {'birth_rates': pd.read_csv(ss.root/'tests/test_data/nigeria_births.csv')}
realistic_death = {'death_rates': pd.read_csv(ss.root/'tests/test_data/nigeria_deaths.csv')}
realistic_fertility = {'fertility_rates': pd.read_csv(ss.root/'tests/test_data/nigeria_asfr.csv')}

series_death = {'death_rates': pd.Series(
            index=[0, 10, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95],
            data=[0.0046355, 0.000776, 0.0014232, 0.0016693, 0.0021449, 0.0028822, 0.0039143, 0.0053676, 0.0082756, 0.01, 0.02, 0.03, 0.04, 0.06, 0.11, 0.15, 0.21, 0.30],
        )}

# births = ss.births(realistic_birth)
pregnancy = ss.Pregnancy(realistic_fertility)
deaths = ss.Deaths(series_death)
births = ss.Births(realistic_birth)
gon = ss.Gonorrhea({'p_death': 0.5, 'initial': 1000})
sim = ss.Sim(people=ppl, demographics=[pregnancy, deaths], diseases=gon, networks=ss.MFNet(), n_years=100)
sim.initialize()
sim.run()

fig, ax = plt.subplots(2, 1)
# ax[0].plot(sim.tivec, sim.results.births.new, label='Births')
ax[0].plot(sim.tivec, sim.results.pregnancy.births, label='Births')
ax[0].plot(sim.tivec, sim.results.deaths.new, label='Deaths')
ax[1].plot(sim.tivec, sim.results.n_alive)

ax[0].set_title('Births and deaths')
ax[1].set_title('Population size')
ax[0].legend()

plt.show()



"""
Static networks from networkx
"""

# %% Imports and settings
import starsim as ss
import matplotlib.pyplot as plt
import networkx as nx

ppl = ss.People(10000)

# This example runs on one static networks + the maternal network
ppl.networks = ss.Networks(
    ss.StaticNet(graph=nx.erdos_renyi_graph, p=0.0001), ss.MaternalNet()
)

hiv = ss.HIV()
hiv.pars['beta'] = {'static': [0.0008, 0.0008], 'maternal': [0.2, 0]}

sim = ss.Sim(people=ppl, demographics=ss.Pregnancy(), diseases=[hiv, ss.Gonorrhea()])
sim.initialize()
sim.run()

plt.figure()
plt.plot(sim.tivec, sim.results.hiv.n_infected)
plt.title('HIV number of infections')
plt.show()



"""
Typical disease dynamics:
Case A.1:
A custom disease module based on a SEIRS model: A->B->C->D->A, where the duration of each state is predefined (attribute).
Interventions and individual behaviours can affect the duration parameters.

Case A.2:
A custom disease module based on a SEIRS model: A->B->C->D->A, where the duration of each state is an output, because the
transition between two states is given by transition probabilities (attribute).
Interventions and individual behaviours can affect the transition probability parameters.

Case A.3: (default structure of all modules).
A custom disease module based on a SEIRS model: A->B->C->D->A, where some transition between two states are predefined,
and other transitions are given by transition probabilities. For instance the transition from A to B is governed by
by a transition probability that depends on network interactions.

Case B:
A custom disease module based on a A.3 model but it's different for females and males:
female:  A->B->C->D->A
male:    A->B->C->F->A

Case C:
A custom disease module based on a nested state machine model, that is there are superstates and substates.
Superstate A:
    Substate B1: B1.1 -> B1.2 -> B1.3
    Substate B2: B2.1 -> B2.2 -> B2.3
Superstate C:
    Substate D1: D1.1 -> D1.2 -> D1.3
    Substate D2: D2.1 -> D2.2 -> D2.3
Where the transitions between substates are governed by predefined durations,
and transitions between superstates are determined by transition probabilities.

"""

# Import STIsimimport stisim as ss
import numpy as np
import stisim as ss
import sciris as sc



class CustomDisease(ss.Module):
    """
    Based on Case A.3:  A->B->C->D->A, the transition from A to B is governed by by a transition probability
    that depends on network interactions.

    Uses custom states to handle duration and transition prob distributions
    """
    def __init__(self, pars=None):
        super().__init__(pars)
        # Define disease states or stages of disease progression, and state trackers
        self.states = ss.omerge(ss.named_dict(
            ss.CustomState('susceptible', bool, True, next_state='infected'),
             self.states)   # Recovered, before becoming susceptible again

        # Initialise additional states to track events
        self.init_event_tracker(['infected'])

    def init_event_trackers(self, states):
        """
        Initialise EventTracker instances
        """
        event_trackers = sc.objdict()
        for state_name in states:
            event_tracker_name = f"ti_{self.states[state_name]}"
            event_trackers[event_tracker_name] = EventTracker(event_tracker_name, fill_value=np.nan)
        self.states = ss.omerge(event_trackers, self.states)

    def next_state_start(self, sim, n, source_state=None):
        """
        Draw len(uids) samples with durations of "source state, to determine the start of "next state".
        """
        ns_start = sim.ti + self.states[source_state].samples(n) / sim.pars.dt
        return ns_start

    def next_state_prob(self, sim, n, source_state=None):
        """
        Draw len(uids) samples with durations of "source state, to determine the start of "next state".
        """
        ns_prob = sim.ti + self.states[source_state].samples(n)
        return ns_prob


class DurationBasedDisease(CustomDisease):
    """
    Case A.1-ish:
    A custom disease module based on a SEIRS model: A->B->C->D->A, where the duration of each state is predefined (attribute).
    Transition from susceptibility to infected is governed by a transition probability that depends on network interaction.
    """
    def __init__(self, pars=None):
        super().__init__(pars)
        # Define disease states or stages of disease progression, and state trackers
        self.states = ss.omerge(ss.named_dict(
            ss.CustomState('susceptible', bool, True, next_state='infected'),
            ss.CStochState('infected',   bool, False, duration=dict(dist='lognormal', par1=0.5, par2=0.5), next_state='infectious'),     # infected
            ss.CStochState('infectious', bool, False, duration=dict(dist='lognormal', par1=0.8, par2=1.2), next_state='recovered'),      # infectious
            ss.CStochState('recovered',  bool, False, duration=dict(dist='lognormal', par1=0.05, par2=0.2), next_state='susceptible')),  # recovered
            self.states)   # Recovered, before becoming susceptible again

        # Initialise additional states to track events
        self.init_event_tracker(['infectious', 'recovered'])

    def set_prognoses(self, sim, uids):
        # Time at which another event will happen (ie, transition from source state to a destination state)
        #infectious_start = sim.ti + np.random.poisson(self.state_durations['infected'] / sim.pars.dt,
        #                                              len(uids[is_infectious]))

        # The assumption here is that the infected state is not infectious, but latent. Infected state has a duration
        # before becoming infectious, and assumes everyone who is infected will become infectious eventually.
        infectious_start = self.next_state_start(sim, len(uids), source_state='infected')

        # Keep track of when events happen
        sim.people[self.name].ti_infectious[uids] = infectious_start




class ProbabilityBasedDisease(ss.Module):
    """
    Case A.2-ish: A custom disease module, where we specify transition probability distributions.
    Transition from susceptibility to infected is governed by a transition probability that depends on network interaction.
    """

    def __init__(self, pars=None):
        super().__init__(pars)
        # Define disease states or stages of disease progression, and state trackers
        self.states = ss.omerge(ss.named_dict(
            ss.CustomState('susceptible', bool, True, next_state='infected'),
            ss.CStochState('infected', bool, False,   transition_prob=dict(dist='uniform', par1=0.0, par2=1.0), next_state='infectious'),  # transition_prob to infectious
            ss.CStochState('infectious', bool, False, transition_prob=dict(dist='normal', par1=0.8, par2=0.1), next_state='recovered'),    # transition_prob to recovered
            ss.CStochState('recovered', bool, False,  transition_prob=1.0, next_state='susceptible')),                                     # transition_prob to susceptible
            self.states)

        # Initialise additional states to track events
        self.init_event_tracker(['infectious', 'recovered'])

    def set_prognoses(self, sim, uids):
        """
        This function updates the disease states (and outcomes) for individuals
        who have newly contracted the disease.
        """
        # Transition from susceptible to infected
        sim.people[self.name].susceptible[uids] = False
        sim.people[self.name].infected[uids] = True
        sim.people[self.name].ti_infected[uids] = sim.ti

        # Proportion of those who are infected decide who will become infectious and who will become latent
        # If transition probabilty is specified as a single number in self.pars
        #is_infectious = np.random.random(len(uids)) < self.get_state_transition_prob('infected', 'infectious')

        # OR if transition probability is specified with custom states
        is_infectious = np.random.random(len(uids)) < self.next_state_prob(sim, len(uids), source_state='infected')

        # Keep track of time events
        sim.people[self.name].ti_infectious[uids[is_infectious]] = sim.ti

        # More prognoses logic here .... check who recovers, etc



# Custom states
class EventTracker(ss.State):
    """
    Convinience class for ti_state States
    """
    def __init__(self, name, dtype=ss.default_float, fill_value=np.nan, **kwargs):
        super().__init__(name, dtype, fill_value, kwargs)
        return


class CustomState(ss.State):
    """
    Custom sate with a hint to the "next state", not meaningfully used yet, just a label
    """
    def __init__(self, name, dtype, fill_value=0, next_state=None, **kwargs):
        super().__init__(name, dtype, fill_value, kwargs)
        self.next_state = next_state
        return

    @property
    def ndim(self):
        return len(sc.tolist(self.shape)) + 1

    def new(self, n):
        shape = sc.tolist(self.shape)
        shape.append(n)
        return np.full(shape, dtype=self.dtype, fill_value=self.fill_value)


class CStochState(ss.State):
    """
    A custom stoachastic state, that returns samples of duration or probabilities
    """
    def __init__(self, name, dtype, duration=None, transition_prob=None, **kwargs):
        super().__init__(name, dtype, **kwargs)
        self.distdict = duration or transition_prob
        return

    def samples(self, n):
        """
        Sample durations or transition probabilities from a distribution
        """
        shape = sc.tolist(self.shape)
        shape.append(n)
        if isinstance(self.distdict, (int, float)):
            return np.full(shape, dtype=self.dtype, fill_value=self.distdict)
        return ss.sample(**self.distdict, size=tuple(shape))



import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from pandas.util import hash_pandas_object
import hashlib
import sciris as sc

from starsim.utils import combine_rands

import warnings
warnings.filterwarnings("ignore", "overflow encountered in scalar subtract")
warnings.filterwarnings("ignore", "overflow encountered in scalar multiply")
 
n = 8 # Number of nodes
n_sources = 2 # Number of sources (seed infections)
reps = 1_000_000
edge_prob = 0.7 # Edge probability
trans_prob = 0.4
seed = 1

def hash(df):
    #return int(hashlib.sha256(hash_pandas_object(df, index=True).values).hexdigest(), 16)
    return hashlib.sha256(hash_pandas_object(df, index=True).values).hexdigest()[:6]

def random(G):
    src = nx.get_node_attributes(G, 'infected')
    infected = []
    tx = []

    # Randomize in case a node is infected multiple times
    el = list(G.edges())
    np.random.shuffle(el)

    for (n1,n2) in el:
        # n1 --> n2
        if src[n1] and (not src[n2]) and (n2 not in infected) and (np.random.rand() < trans_prob):
            tx.append((n1, n2))
            infected.append(n2)
        # n2 --> n1
        if src[n2] and (not src[n1]) and (n1 not in infected) and (np.random.rand() < trans_prob):
            tx.append((n2, n1))
            infected.append(n1)
    return tx


def modulo(G):
    src = nx.get_node_attributes(G, 'infected')
    infected = []
    tx = []
    r1 = np.random.rand(n)
    r2 = np.random.rand(n)
    el = list(G.edges())
    np.random.shuffle(el)
    for (n1,n2) in el:
        # n1 --> n2
        if (src[n1]) and (not src[n2]) and (n2 not in infected) and (((r1[n1]+r2[n2])%1) < trans_prob):
            tx.append((n1, n2))
            infected.append(n2)
        # n2 --> n1
        if (src[n2]) and (not src[n1]) and (n1 not in infected) and (((r1[n2]+r2[n1])%1) < trans_prob):
            tx.append((n2, n1))
            infected.append(n1)
    return tx


def combine_hash(G):
    src = nx.get_node_attributes(G, 'infected')
    infected = []
    tx = []
    el = list(G.edges())
    r1 = np.random.randint(low=0, high=np.iinfo(np.uint64).max, dtype=np.uint64, size=n)
    r2 = np.random.randint(low=0, high=np.iinfo(np.uint64).max, dtype=np.uint64, size=n)
    np.random.shuffle(el)
    for (n1,n2) in el:
        # n1 --> n2
        r = combine_rands(r1[n1], r2[n2])
        if (src[n1]) and (not src[n2]) and (n2 not in infected) and (r < trans_prob):
            tx.append((n1, n2))
            infected.append(n2)
        # n2 --> n1
        r = combine_rands(r1[n2], r2[n1])
        if (src[n2]) and (not src[n1]) and (n1 not in infected) and (r < trans_prob):
            tx.append((n2, n1))
            infected.append(n1)
    return tx

def roulette(G):
    src = nx.get_node_attributes(G, 'infected')
    tx = []
    for n2 in G.nodes():
        # All nodes --> n2
        if not src[n2]:
            srcs = [n for n in G.neighbors(n2) if src[n]] #[src[n] for n in G.neighbors(n2)] # Count infected neighbors
            cnt = len(srcs)
            if np.random.rand() < 1-(1-trans_prob)**cnt:
                n1 = np.random.choice(srcs)
                tx.append((n1, n2))
    return tx
 

def transmit(G, trans_fn):
    # Align transmissions from tx_in if provided
    txs = {}
    counts = {}
 
    for _ in np.arange(reps):
        txl = trans_fn(G)
        tx = pd.DataFrame(txl, columns=['src', 'dst']).sort_values(['src', 'dst']).reset_index(drop=True)
        
        h = hash(tx)

        if h not in txs:
            txs[h] = tx

        counts[h] = counts.get(h, 0) + 1

    df = pd.DataFrame(counts.values(), index=pd.Index(counts.keys(), name='Hash'), columns=['Counts'])
    return txs, df


# Build the graph
G = nx.random_graphs.erdos_renyi_graph(n=n, p=edge_prob, seed=seed)

# Seed infections
infected = {i:False for i in range(n)}
sources = np.random.choice(a=range(n), size=n_sources, replace=False)
for source in sources:
    infected[source] = True
nx.set_node_attributes(G, infected, 'infected')
 
# Do transmissions via each method in parallel
results = sc.parallelize(transmit, iterkwargs=[{'trans_fn':random}, {'trans_fn':modulo}, {'trans_fn':combine_hash}, {'trans_fn':roulette}], kwargs={'G':G}, die=True, serial=False)
tx, cnt = zip(*results)

df = pd.concat(cnt, axis=1) \
    .fillna(0) \
    .astype(int)
df.columns = ['Random', 'Modulo', 'CombineHash', 'Roulette']

# Manipulate results
df.reset_index(inplace=True)
dfm = df.melt(id_vars='Hash', var_name='Method', value_name='Count')

# Plot
g = sns.barplot(data=dfm, x='Hash', y='Count', hue='Method')
plt.xticks(rotation=90)
g.figure.tight_layout()

txc = tx[0].copy()
for i in range(1, len(tx)):
    txc.update(tx[i])
for h, v in txc.items():
    print(f'\nUnique transmission tree #{h}')
    print(v)

plt.figure()
pos = nx.spring_layout(G, seed=3113794652)  # positions for all nodes
ec = nx.draw_networkx_edges(G, pos, alpha=0.2)
colors = ['red' if infected else 'blue' for infected in nx.get_node_attributes(G, 'infected').values()]
nc = nx.draw_networkx_nodes(G, pos, nodelist=G.nodes(), node_color=colors, node_size=100)
nx.draw_networkx_edges(G, pos, width=1.0, alpha=0.5)
nx.draw_networkx_labels(G, pos)

plt.show()


"""
Parent class for the integration loop.
"""
import time
import numpy as np
import pandas as pd
import sciris as sc
import starsim as ss
import matplotlib.pyplot as plt

# What classes are externally visible
__all__ = ['Loop']



#%% Loop class

class Loop:
    """ Base class for integration loop """
    def __init__(self, sim): # TODO: consider eps=1e-6 and round times to this value
        self.sim = sim
        self.funcs = None
        self.abs_tvecs = None
        self.plan = None
        self.index = 0 # The next function to execute
        self.cpu_time = [] # Store the CPU time of execution of each function
        self.df = None # User-friendly verison of the plan
        self.cpu_df = None # User-friendly time analysis
        self.initialized = False
        return

    def init(self):
        """ Parse the sim into the integration plan """
        self.collect_funcs()
        self.collect_abs_tvecs()
        self.make_plan()
        self.initialized = True
        return

    def __len__(self):
        if self.initialized:
            return len(self.plan)
        else:
            return 0 # Or None?

    def __iadd__(self, func):
        """ Allow functions to be added to the function list """
        parent = func.__self__
        func_name = func.__name__

        # Get the name if it's defined, the class otherwise; these must match abs_tvecs
        module = parent.name if isinstance(parent, ss.Module) else parent.__class__.__name__.lower()

        # Create the row and append it to the function list
        row = dict(func_order=len(self.funcs), module=module, func_name=func_name, func=func)
        self.funcs.append(row)
        return self

    def __repr__(self):
        if self.initialized:
            arrs = list({len(arr) for arr in self.abs_tvecs.values()})
            if len(arrs) == 1: arrs = arrs[0] # If all are the same, just use that
            string = f'Loop(n={len(self)}, funcs={len(self.funcs)}, npts={arrs}, index={self.index})'
        else:
            string = 'Loop(initialized=False)'
        return string

    def disp(self):
        return sc.pr(self)

    def collect_funcs(self):
        """ Collect all the callable functions (methods) that comprise the step """

        # Run the simulation step first (updates the distributions)
        self.funcs = [] # Reset, just in case
        sim = self.sim

        # Collect the start_steps
        self += sim.start_step # Note special __iadd__() method above, which appends these to the funcs list
        for mod in sim.modules:
            self += mod.start_step

        # Update demographic modules (create new agents from births/immigration, schedule non-disease deaths and emigration)
        for dem in sim.demographics():
            self += dem.step

        # Carry out autonomous state changes in the disease modules. This allows autonomous state changes/initializations
        # to be applied to newly created agents
        for disease in sim.diseases():
            if isinstance(disease, ss.Disease): # Could be a connector instead -- TODO, rethink this
                self += disease.step_state

        # Update connectors
        for connector in sim.connectors():
            self += connector.step

        # Update networks - this takes place here in case autonomous state changes at this timestep
        # affect eligibility for contacts
        for network in sim.networks():
            self += network.step

        # Apply interventions - new changes to contacts will be visible and so the final networks can be customized by
        # interventions, by running them at this point
        for intv in sim.interventions():
            self += intv.step

        # Carry out autonomous state changes in the disease modules, including transmission (but excluding deaths)
        for disease in sim.diseases():
            self += disease.step

        # Update people who died -- calls disease.step_die() internally
        self += sim.people.step_die

        # Update results
        self += sim.people.update_results
        for mod in sim.modules:
            self += mod.update_results

        # Apply analyzers
        for ana in sim.analyzers():
            self += ana.step

        # Clean up dead agents, increment the time index, and perform other housekeeping tasks
        for mod in sim.modules:
            self += mod.finish_step
        self += sim.people.finish_step
        self += sim.finish_step

        return self.funcs

    def collect_abs_tvecs(self):
        """ Collect numerical time arrays for each module """
        self.abs_tvecs = sc.objdict()

        # Handle the sim and people first
        sim = self.sim
        for key in ['sim', 'people']:
            self.abs_tvecs[key] = sim.t.abstvec

        # Handle all other modules
        for mod in sim.modules:
            self.abs_tvecs[mod.name] = mod.t.abstvec

        return self.abs_tvecs

    def make_plan(self):
        """ Combine the module ordering and the time vectors into the integration plan """

        # Assemble the list of dicts
        raw = []
        for func_row in self.funcs:
            for t in self.abs_tvecs[func_row['module']]:
                row = func_row.copy()
                row['time'] = t # Add time column
                raw.append(row)

        # Turn it into a dataframe
        self.plan = sc.dataframe(raw)

        # Sort it by step_order, a combination of time and function order
        self.plan['step_order'] = self.plan.time + ss.options.time_eps*self.plan.func_order
        self.plan['func_label'] = self.plan.module + '.' + self.plan.func_name
        col_order = ['time', 'func_order', 'step_order', 'func', 'func_label', 'module', 'func_name'] # Func in the middle to hide it
        self.plan = self.plan.sort_values('step_order').reset_index(drop=True)[col_order]
        return

    def store_time(self):
        """ Store the current time in as high resolution as possible """
        self.cpu_time.append(time.perf_counter())
        return

    def run_one_step(self):
        """
        Take a single step, i.e. call a single function; only used for debugging purposes.

        Compare sim.run_one_step(), which runs a full timestep (which involves multiple function calls).
        """
        f = self.plan.func[self.index] # Get the next function
        f() # Call it
        self.index += 1 # Increment the time
        return

    def run(self, until=None, verbose=None):
        """ Actually run the integration loop; usually called by sim.run() """

        # Convert e.g. '2020-01-01' to an actual date
        if isinstance(until, str):
            until = ss.date(until)

        # Loop over every function in the integration loop, e.g. disease.step()
        self.store_time()
        for f,label in zip(self.plan.func[self.index:], self.plan.func_label[self.index:]):
            if verbose:
                row = self.plan[self.index]
                print(f'Running t={row.time:n}, step={row.step_order}, {label}()')

            f() # Execute the function -- this is where all of Starsim happens!!

            # Tidy up
            self.index += 1 # Increment the count
            self.store_time()
            if until is not None and self.sim.now > until: # Terminate if asked to
                break
        return

    def manual_reset(self): # TODO: do we need this? I feel if we don't have it, people will be tempted to manually set loop.index = 0.
        """
        Reset the loop to run again. Note, does not reset sim quantities so should
        only be used for debugging.
        """
        self.index = 0
        self.sim.complete = False
        return

    def to_df(self):
        """ Return a user-friendly version of the plan, omitting object columns """
        # Compute the main dataframe
        cols = ['time', 'func_order', 'module', 'func_name', 'func_label']
        if self.plan is not None:
            df = self.plan[cols].copy() # Need to copy, otherwise it's messed up
        else:
            errormsg = f'Simulation "{self.sim}" needs to be initialized before exporting the Loop dataframe'
            raise RuntimeError(errormsg)
        times = np.diff(self.cpu_time)
        if len(times) == len(df):
            df['cpu_time'] = times
        else:
            df['cpu_time'] = np.nan
        self.df = df

        # Compute the CPU dataframe
        by_func = df.groupby('func_label')
        method = dict(func_order='first', module='first', func_name='first', cpu_time='sum')
        cdf = sc.dataframe(by_func.agg(method))
        cdf['percent'] = cdf.cpu_time / cdf.cpu_time.sum()*100
        cdf.insert(cdf.cols.index('cpu_time'), 'calls', by_func.size())
        cdf.sort_values('cpu_time', inplace=True, ascending=False)
        self.cpu_df = cdf
        return df

    def plot(self, simplify=False, fig_kw=None, plot_kw=None, scatter_kw=None):
        """
        Plot a diagram of all the events

        Args:
            simplify (bool): if True, skip update_results and finish_step events, which are automatically applied
            fig_kw (dict): passed to ``plt.figure()``
            plot_kw (dict): passed to ``plt.plot()``
            scatter_kw (dict): passed to ``plt.scatter()``
        """

        # Assemble data
        df = self.to_df()
        if simplify:
            filter_out = ['update_results', 'finish_step']
            df = df[~df.func_name.isin(filter_out)]
        yticks = df.func_order.unique()
        ylabels = df.func_label.unique()
        x = df.time
        y = df.func_order

        # Convert module names to integers for plotting colors
        mod_int, _ = pd.factorize(df.module)
        colors = sc.gridcolors(np.unique(mod_int), asarray=True)

        # Do the plotting
        plot_kw = sc.mergedicts(dict(lw=2, alpha=0.2, c='k'), plot_kw)
        scatter_kw = sc.mergedicts(dict(s=200, alpha=0.6), scatter_kw)
        fig = plt.figure(**sc.mergedicts(fig_kw))
        plt.plot(x, y, **plot_kw)
        plt.scatter(x, y, c=colors[mod_int], **scatter_kw)
        plt.yticks(yticks, ylabels)
        plt.title(f'Integration plan ({len(df)} events)')
        plt.xlabel(f'Time since simulation start (in {self.sim.pars.unit}s)')
        plt.grid(True)
        sc.figlayout()
        sc.boxoff()
        return ss.return_fig(fig)

    def plot_cpu(self, bytime=True, fig_kw=None, bar_kw=None):
        """
        Plot the CPU time spent on each event; visualization of Loop.cpu_df.

        Args:
            bytime (bool): if True, order events by total time rather than actual order
            fig_kw (dict): passed to ``plt.figure()``
            plot_kw (dict): passed to ``plt.bar()``
        """
        # Assemble data
        if self.cpu_df is None:
            self.to_df()
        df = self.cpu_df
        ylabels = df.index.values
        if bytime:
            y = np.arange(len(ylabels))
        else:
            y = df.func_order.values
        y = y[::-1] # Reverse order so plots from top to bottom

        x = df.cpu_time.values
        pcts = df.percent.values

        if x.max() < 1:
            x *= 1e3
            unit = 'ms'
        else:
            unit = 's'

        # Assemble labels
        for i in range(len(df)):
            timestr = sc.sigfig(x[i], 3) + f' {unit}'
            pctstr = sc.sigfig(pcts[i], 3) + '%'
            ylabels[i] += f'()\n{timestr}, {pctstr}'

        # Do the plotting
        bar_kw = sc.mergedicts(bar_kw)
        fig = plt.figure(**sc.mergedicts(fig_kw))
        plt.barh(y, width=x, **bar_kw)
        plt.yticks(y, ylabels)
        plt.xlabel(f'CPU time ({unit})')
        plt.ylabel('Function call')
        plt.grid(True)
        sc.figlayout()
        sc.boxoff()
        return ss.return_fig(fig)

    def __deepcopy__(self, memo):
        """ A dataframe that has functions in it doesn't copy well; convert to a dict first """
        cls = self.__class__
        new = cls.__new__(cls)
        memo[id(self)] = new
        for k, v in self.__dict__.items():
            if k == 'plan' and isinstance(v, sc.dataframe):
                origdict = v.to_dict() # Convert to a dictionary
                newdict = sc.dcp(origdict, memo=memo) # Copy the dict
                newdf = sc.dataframe(newdict)
                setattr(new, k, newdf)
            else:
                setattr(new, k, sc.dcp(v, memo=memo)) # Regular deepcopy
        return new




"""
Test births and deaths
"""

# %% Imports and settings
import starsim as ss
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import sciris as sc

class agent_analyzer(ss.Analyzer):
    def init_pre(self, sim):
        super().init_pre(sim)
        self.n_agents = np.zeros(sim.npts)

    def update_results(self, sim):
        self.n_agents[sim.ti] = len(sim.people)


def run_test(remove_dead=True, do_plot=False, rand_seed=0, verbose=False):
    ppl = ss.People(50000)

    # Parameters
    realistic_birth = {'birth_rate': pd.read_csv(ss.root/'tests/test_data/nigeria_births.csv')}

    series_death = {
        'death_rate': pd.Series(
            index=[0, 10, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95],
            data=[0.0046355, 0.000776, 0.0014232, 0.0016693, 0.0021449, 0.0028822, 0.0039143, 0.0053676, 0.0082756, 0.01, 0.02, 0.03, 0.04, 0.06, 0.11, 0.15, 0.21, 0.30],
        ),
        'rel_death': 7
    }

    births = ss.Births(realistic_birth)
    deaths = ss.Deaths(series_death)
    gon = ss.Gonorrhea({'p_death': 0.5, 'init_prev': 0.1})

    sim = ss.Sim(people=ppl, demographics=[births,deaths], diseases=[gon], networks=ss.MFNet(), remove_dead=remove_dead, n_years=100, analyzers=[agent_analyzer()], rand_seed=rand_seed, verbose=verbose)

    sim.initialize()
    sim.run()

    if do_plot:
        fig, ax = plt.subplots(2, 1)
        ax[0].plot(sim.tivec, sim.results.births.new, label='Births')
        ax[0].plot(sim.tivec, sim.results.background_deaths.new, label='Deaths')
        ax[1].plot(sim.tivec, sim.results.n_alive, label='Number of alive agents')
        ax[1].plot(sim.tivec, sim.analyzers[0].n_agents, label='Number of agents')
        ax[0].set_title('Births and deaths')
        ax[1].set_title('Population size')
        ax[0].legend()
        ax[1].legend()
        fig.suptitle(f'Remove people {remove_dead}')
        fig.tight_layout()
        plt.show()

    return sim


def compare_totals(rand_seed=0):
    sim1 = run_test(False, rand_seed=rand_seed)
    sim2 = run_test(True, rand_seed=rand_seed+1)
    return sim1.people.alive.count_nonzero() - sim2.people.alive.count_nonzero()


if __name__ == '__main__':

    # x = sc.parallelize(compare_totals, np.arange(50))

    sim1 = run_test(False, rand_seed=0, verbose=0.1)
    sim2 = run_test(True, rand_seed=0, verbose=0.1)

    fig, ax = plt.subplots(2, 2)
    ax = ax.ravel()

    ax[0].plot(sim1.tivec, sim1.results.births.new, label='No removal')
    ax[0].plot(sim2.tivec, sim2.results.births.new, label='Removal', linestyle='--')
    ax[0].set_title('Births')


    ax[1].plot(sim1.tivec, sim1.results.deaths.new, label='No removal')
    ax[1].plot(sim2.tivec, sim2.results.deaths.new, label='Removal', linestyle='--')
    ax[1].set_title('Background deaths')

    ax[1].plot(sim1.tivec, sim1.results.new_deaths, label='(all) No removal')
    ax[1].plot(sim2.tivec, sim2.results.new_deaths, label='(all) Removal', linestyle='--')


    ax[2].plot(sim1.tivec, sim1.results.n_alive, label='No removal')
    ax[2].plot(sim2.tivec, sim2.results.n_alive, label='Removal', linestyle='--')
    ax[2].set_title('Number of alive agents')


    ax[3].plot(sim1.tivec, sim1.analyzers[0].n_agents, label='No removal')
    ax[3].plot(sim2.tivec, sim2.analyzers[0].n_agents, label='Removal', linestyle='--')
    ax[3].set_title('Number of agents')

    fig.tight_layout()
    plt.show()



"""
Test calibration
"""

#%% Imports and settings
import sciris as sc
import starsim as ss
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

#from ax.plot.contour import plot_contour
#from ax.plot.trace import optimization_trace_single_method
#from ax.service.managed_loop import optimize
#from ax.utils.notebook.plotting import init_notebook_plotting, render

from ax.service.ax_client import AxClient, ObjectiveProperties
from ax.utils.notebook.plotting import init_notebook_plotting, render

from ax.modelbridge.cross_validation import cross_validate
from ax.plot.contour import interact_contour
from ax.plot.diagnostic import interact_cross_validation
from ax.plot.scatter import interact_fitted, plot_objective_vs_constraints, tile_fitted
from ax.plot.slice import plot_slice
from ax.service.utils.report_utils import exp_to_df

do_plot = 1
do_save = 0
n_agents = [2e3, 25_000][1]

ax_client = AxClient(enforce_sequential_optimization=False)

#%% Helper functions

def make_sim(calib_pars):
    sir = ss.SIR(
        beta = ss.beta( calib_pars.get('beta', 0.9) ),
        dur_inf = ss.lognorm_ex(mean=ss.dur( calib_pars.get('dur_inf', 6))),
        init_prev = ss.bernoulli(0.01),
    )

    #deaths = ss.Deaths(death_rate=15)
    #births = ss.Births(birth_rate=15)

    random = ss.RandomNet(n_contacts=ss.poisson(calib_pars.get('n_contacts', 4)))

    sim = ss.Sim(
        dt = 1,
        unit = 'day',
        n_agents = n_agents,
        #total_pop = 9980999,
        start = sc.date('2024-01-01'),
        stop = sc.date('2024-01-31'),
        diseases = sir,
        networks = random,
        #demographics = [deaths, births],
        rand_seed = np.random.randint(1e6),
    )

    return sim


def eval_sim(pars):
    sim = make_sim(pars)
    sim.run()

    if False:
        fig = sim.plot()
        fig.suptitle(pars)
        fig.subplots_adjust(top=0.9)
        plt.show()

    return dict(
        prevalence_error = (np.abs(sim.results.sir.prevalence[-1] - 0.20), None),
        #prevalence = (sim.results.sir.prevalence[-1], None),
    )

#%% Define the tests
def test_calibration(do_plot=False):
    sc.heading('Testing calibration')

    # Define the calibration parameters
    calib_pars = [
        dict(name='beta', type='range', bounds=[0.005, 0.1], value_type='float', log_scale=True),
        #dict(name='dur_inf', type='range', bounds=[1, 120], value_type='float', log_scale=False),
        dict(name='dur_inf', type='fixed', value=60, value_type='float'),
        #dict(name='init_prev', type='range', bounds=[0.01, 0.30], value_type='float', log_scale=False),
        dict(name='n_contacts', type='range', bounds=[1, 10], value_type='int', log_scale=False),
    ]

    ax_client.create_experiment(
        name = 'starsim test',
        parameters = calib_pars,
        objectives={'prevalence_error': ObjectiveProperties(minimize=True)},
        parameter_constraints = None,
        outcome_constraints = None,
        choose_generation_strategy_kwargs={"max_parallelism_override": 25},
    )

    print('Max parallelism:', ax_client.get_max_parallelism()) # Seems to require manual specification of generation_strategy

    for i in range(5):
        print('THINKING...')
        trial_index_to_param, idk = ax_client.get_next_trials(max_trials=1_000)

        print('STEP', i, len(trial_index_to_param))

        # Does NOT work to complete_trial in the parallel loop
        results = sc.parallelize(eval_sim, iterkwargs=dict(pars=list(trial_index_to_param.values())), serial=False)
        for trial_index, result in zip(trial_index_to_param.keys(), results):
            ax_client.complete_trial(trial_index=trial_index, raw_data=result)

        print(exp_to_df(ax_client.experiment))


    best_pars, values = ax_client.get_best_parameters()

    return best_pars, values#, exp, model


#%% Run as a script
if __name__ == '__main__':


    T = sc.timer()
    do_plot = True

    best_pars, values = test_calibration(do_plot=do_plot)

    sim = make_sim(best_pars)
    sim.run()
    sim.plot()

    print('best_pars:', best_pars)
    print('values:', values)

    #render(ax_client.get_contour_plot(param_x='beta', param_y='dur_inf', metric_name='prevalence_error'))
    render(ax_client.get_optimization_trace(objective_optimum=0))

    model = ax_client.generation_strategy.model
    render(interact_contour(model=model, metric_name='prevalence_error'))

    cv_results = cross_validate(model)
    render(interact_cross_validation(cv_results))

    render(plot_slice(model, 'beta', 'prevalence_error'))
    render(plot_slice(model, 'n_contacts', 'prevalence_error'))

    render(interact_fitted(model, rel=False))

    plt.show()

    T.toc()



"""
A simple beta sweep with multisim on and off using HIV as an example pathogen
"""

# %% Imports and settings
import os
import starsim as ss
import scipy.stats as sps
import sciris as sc
import pandas as pd
import seaborn as sns

# Suppress warning from seaborn
import warnings
warnings.filterwarnings("ignore", "is_categorical_dtype")
warnings.filterwarnings("ignore", "use_inf_as_na")

n = 1_000 # Agents
n_rand_seeds = 25
xf_levels = [0.5, 0.8, 1.26, 2.0] + [1] # Must include 1 as that's the baseline | roughly np.logspace(np.log2(0.5), np.log2(20), 4, base=2)

figdir = os.path.join(os.getcwd(), 'figs', 'Sweep')
sc.path(figdir).mkdir(parents=True, exist_ok=True)

def run_sim(n, xf, rand_seed, multirng):
    ppl = ss.People(n)

    rel_pars = {
        'male_shift': 5,
        'std': 3,
        'pars': {
            'dur': sps.lognorm(15*xf, 15*xf),  # Can vary by age, year, and individual pair
            'part_rates': 0.9,  # Participation rates - can vary by sex and year
            'rel_part_rates': 1.0,
            'debut': 16,  # Age of debut can vary by sex, year, and individual
        }} 
    ppl.networks = ss.ndict(ss.EmbeddingNet(**rel_pars), ss.MaternalNet())

    hiv_pars = {
        #'beta': {'embedding': [xf * 0.30, xf * 0.25], 'maternal': [xf * 0.2, 0]},
        'beta': {'embedding': [0.30, 0.25], 'maternal': [0.2, 0]},
        'init_prev': 0.05
    }
    hiv = ss.HIV(hiv_pars)

    pregnancy = ss.Pregnancy()

    pars = {
        'start': 1980,
        'end': 2020,
        'rand_seed': rand_seed,
    }
    sim = ss.Sim(people=ppl, diseases=[hiv], demographics=[pregnancy], pars=pars, label=f'Sim with {n} agents and xf={xf}')
    sim.initialize()
    sim.run()

    df = pd.DataFrame( {
        'ti': sim.tivec,
        #'hiv.n_infected': sim.results.hiv.n_infected,
        'hiv.prevalence': sim.results.hiv.prevalence,
        'hiv.cum_deaths': sim.results.hiv.new_deaths.cumsum(),
        'pregnancy.cum_births': sim.results.pregnancy.births.cumsum(),
    })
    df['xf'] = xf
    df['rand_seed'] = rand_seed
    df['multirng'] = multirng

    return df

def run_scenarios(figdir):
    results = []
    times = {}
    for multirng in [True, False]:
        ss.options(multirng=multirng)
        cfgs = []
        for rs in range(n_rand_seeds):
            for xf in xf_levels:
                cfgs.append({'xf':xf, 'rand_seed':rs, 'multirng':multirng})
        T = sc.tic()
        results += sc.parallelize(run_sim, kwargs={'n': n}, iterkwargs=cfgs, die=True)
        times[f'multirng={multirng}'] = sc.toc(T, output=True)

    print('Timings:', times)

    df = pd.concat(results)
    df.to_csv(os.path.join(figdir, 'result.csv'))
    return df


def plot_scenarios(df, figdir):
    d = pd.melt(df, id_vars=['ti', 'rand_seed', 'xf', 'multirng'], var_name='channel', value_name='Value')
    d['baseline'] = d['xf']==1
    bl = d.loc[d['baseline']]
    scn = d.loc[~d['baseline']]
    bl = bl.set_index(['ti', 'channel', 'rand_seed', 'xf', 'multirng'])[['Value']].reset_index('xf')
    scn = scn.set_index(['ti', 'channel', 'rand_seed', 'xf', 'multirng'])[['Value']].reset_index('xf')
    mrg = scn.merge(bl, on=['ti', 'channel', 'rand_seed', 'multirng'], suffixes=('', '_ref'))
    mrg['Value - Reference'] = mrg['Value'] - mrg['Value_ref']
    mrg = mrg.sort_index()

    fkw = {'sharey': False, 'sharex': 'col', 'margin_titles': True}

    ## TIMESERIES
    g = sns.relplot(kind='line', data=d, x='ti', y='Value', hue='xf', col='channel', row='multirng',
        height=5, aspect=1.2, palette='Set1', errorbar='sd', lw=2, facet_kws=fkw)
    g.set_titles(col_template='{col_name}', row_template='multirng: {row_name}')
    g.set_xlabels(r'$t_i$')
    g.fig.savefig(os.path.join(figdir, 'timeseries.png'), bbox_inches='tight', dpi=300)

    ## DIFF TIMESERIES
    for ms, mrg_by_ms in mrg.groupby('multirng'):
        g = sns.relplot(kind='line', data=mrg_by_ms, x='ti', y='Value - Reference', hue='xf', col='channel', row='xf',
            height=3, aspect=1.0, palette='Set1', estimator=None, units='rand_seed', lw=0.5, facet_kws=fkw) #errorbar='sd', lw=2, 
        g.set_titles(col_template='{col_name}', row_template='Beta: {row_name}')
        g.fig.suptitle('MultiRNG' if ms else 'SingleRNG')
        g.fig.subplots_adjust(top=0.88)
        g.set_xlabels(r'Value - Reference at $t_i$')
        g.fig.savefig(os.path.join(figdir, 'diff_multi.png' if ms else 'diff_single.png'), bbox_inches='tight', dpi=300)

    ## FINAL TIME
    tf = df['ti'].max()
    mtf = mrg.loc[tf]
    g = sns.displot(data=mtf.reset_index(), kind='kde', fill=True, rug=True, cut=0, hue='xf', x='Value - Reference', col='channel', row='multirng',
        height=5, aspect=1.2, facet_kws=fkw, palette='Set1')
    g.set_titles(col_template='{col_name}', row_template='multirng: {row_name}')
    g.set_xlabels(f'Value - Reference at $t_i={{{tf}}}$')
    g.fig.savefig(os.path.join(figdir, 'final.png'), bbox_inches='tight', dpi=300)

    ## FINAL TIME function of beta
    dtf = d.set_index(['ti', 'rand_seed']).sort_index().loc[tf]
    g = sns.relplot(kind='line', data=dtf.reset_index(), x='xf', y='Value', col='channel', row='multirng',
        height=5, aspect=1.2, facet_kws=fkw, estimator=None, units='rand_seed', lw=0.25)
    g.set_titles(col_template='{col_name}', row_template='multirng: {row_name}')
    g.set_ylabels(f'Value at $t_i={{{tf}}}$')
    g.fig.savefig(os.path.join(figdir, 'final_beta.png'), bbox_inches='tight', dpi=300)

    print('Figures saved to:', os.path.join(os.getcwd(), figdir))

    return


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--plot', help='Folder containing cached results.csv', type=str)
    args = parser.parse_args()

    if args.plot:
        print('Reading CSV file', args.plot)
        df = pd.read_csv(os.path.join(args.plot, 'result.csv'), index_col=0)
        figdir = args.plot
    else:
        print('Running scenarios')
        df = run_scenarios(figdir)

    plot_scenarios(df, figdir)

    print('Done')


"""
Vignette 05: Scenarios
The implementations below have some features in common, but each focuses on a different aspect
    - RA focus: iterating over par values
    - PSL focus: branching a sim -- running to a certain point, then brancing into multiple scenarios
    - RS focus: replication of covasim behavior

NB, all of us agreed that we don't use scenarios much. Possibly this should be deprioritized?
"""
import stisim as ss

##################
# @RomeshA
##################
# This is presumably not the user-friendly way of doing it, but is how I would envisage writing my own scripts

def scen_sim(scen_par=value, seed=0):
    ... # apply config e.g., in modules, or in people
    sim = ss.Sim()
    sim.run()
    return sim

sim1 = scen_sim(beta=0.1)
sim2 = scen_sim(beta=0.2)
sim3 = scen_sim(beta=0.3)

# Something more standard?
scen = ss.Scenario(sim, 'hpv.eff_condoms'=[0.1,0.2,0.3])
sims = scen.run()
# But in this setup I'm not sure when the parameter value is consumed. It would have to be after
# the module has been instantiated, but what if the module's `__init__` method did something that depended
# on the parameter value? Would it be too late to apply the scenario? Does the Scenario instead need a sim generation
# function? (that would be too complex for the simple UI)



##################
# @pausz
##################


##################################### Case 1: Branching using multisims ################################################
# Step 0: Run multiple simulations with different seeds
pars = {'start': 1950,
        'end': 2022}
sim = ss.Sim(pars=pars, modules=ss.hiv())
msim = ss.MultiSim(sim)
msim.run(n_runs=100)

# Step 1: estimate average behaviour
median_sim = msim.median()
# Save last 10 years of data to disk, doesn't modify median_sim
median_sim.save(horizon=10)

# Step 2a: use the median sim as history for the next simulation, with a horizon of 10 years (ie, use the last 10 years of
# the "median simulation" instead of using 'burnin' years to avoid trnasient dyanamics).
# This new sim will have to set the "transplanted state": ie, initialise the new simulation's state using the saved
# state from the source simulation. This involves (re)setting the population's attributes, disease states, and any other
# relevant parameters to match the saved state.
pars = {'end': 2100.0,
        'horizon': 10.0}
s1 = ss.Sim(history=median_sim, pars=pars,  modules=ss.syphilis(), label='Syphilis outbreak')

# OR we could do
# Discard everything older than the last 10 years of simulation
median_sim.keep(horizon=10.0)
# Step 2b: transplant history
s1 = ss.Sim(history=median_sim, modules=ss.syphilis(), label='Syphilis outbreak')

# Step 3: Progress forward in time from the end of median_sim onward,
# introduce changes to explore alternative scenarios (branching).
rel_betas = [0.9, 1.0]
msims = []
for beta in rel_betas:
    s1.initialize()
    s1.pars.syphilis.rel_beta = beta
    msim = ss.MultiSim(s1)
    msim.run(n_runs=20)
    msim.mean()
    msims.append(msim)

merged = ss.MultiSim.merge(msims)
merged.plot(color_by_sim=True)

##################################### Case 2: Branching using scenarios ################################################
# Step 0: Define interventions
campaign_a = ss.ther_vaccine(eligibility={'female': {'min_age': 30, 'max_age': 55}}, prob=0.8, start_year=2020)
campaign_b = ss.prev_vaccine(eligibility={'female': {'min_age': 15, 'max_age': 25},
                                          'male': {'min_age': 10, 'max_age': 20}}, prob=0.8, start_year=2010)

screening_campaign = ss.screening(prob=0.2, start_year=2010)

# Step 1: Define parameters that are common to all simulations
basepars = {'beta': [0.9],
            'interventions': [screening_campaign]}

# Step 2: Define specific configurations for scenarios
scenarios = {'campaign_a': {
    'name': 'Therapeutic Campaign',
    'pars': {
        'interventions': [campaign_a]
    }
},
    'campaign_b': {
        'name': 'Preventative Campaign',
        'pars': {
            'interventions': [campaign_b],
        }
    },
}

# Step 3: Each scenario branches from median_sim
scens = ss.Scenarios(history_sim=median_sim, basepars=basepars, scenarios=scenarios)
scens.run()


##################
# @robynstuart
##################
import stisim as ss
import pandas as pd

# Case 1: multisim with stochastic variation
sim = ss.Sim(modules=ss.hiv())
msim = ss.MultiSim(sim)
msim.run(n_runs=5)
msim.plot()

# Case 2: average over stochastic variation
msim.mean()
msim.plot_result('hiv.new_infections')

# Case 3: run simple scenarios
s1 = ss.Sim(modules=ss.syphilis(), label='Without HIV')
s2 = ss.Sim(modules=[ss.hiv(), ss.syphilis()], label='With HIV')
ss.parallel(s1, s2).plot(['syphilis.cum_deaths', 'syphilis.cum_infections'])


# Case 4: run combination msims & scenarios
module_dict = {'Without HIV': ss.syphilis(), 'With HIV': [ss.hiv(), ss.syphilis()]}
syph_rel_betas = [0.9, 1, 1.1]

msims = []
for label, modules in module_dict.items():
    sims = []
    for beta in syph_rel_betas:
        sim = ss.Sim(modules=modules, label=label)
        sim.initialize()
        sim.pars.syphilis.rel_beta = beta  # Could also allow this to be supplied directly
        sims.append(sim)
    msim = ss.MultiSim(sims)
    msim.run()
    msim.mean()
    msims.append(msim)

merged = ss.MultiSim.merge(msims)
merged.plot(color_by_sim=True)


# Case 5: scenarios
basepars = {'modules': [ss.hiv(), ss.syphilis]}

products_dx = pd.read_csv('products_dx.csv')
products_tx = pd.read_csv('products_tx.csv')
syph_dx = ss.Product(efficacy=products_dx[products_dx.name == 'syph_dx'])
syph_tx = ss.Product(efficacy=products_tx[products_tx.name == 'syph_tx'])
syph_screening = ss.screening(products=syph_dx, prob=0.3, start_year=2015)
syph_treatment = ss.treatment(eligibility='syph.diagnosed', products=syph_tx, prob=0.8, start_year=2015)

# Settings for each scenario
scenarios = {'baseline': {
              'name': 'Baseline',
              'pars': {}
              },
            'with_treatment': {
              'name': 'With treatment',
              'pars': {
                  'interventions': [syph_screening, syph_treatment],
                  }
              },
             }

# Run and plot the scenarios
scens = ss.Scenarios(basepars=basepars, scenarios=scenarios)
scens.run()
scens.plot()





"""
Functions and classes for handling time
"""
import sciris as sc
import numpy as np
import pandas as pd
import datetime as dt
import dateutil as du
import starsim as ss

__all__ = ['Date', 'Dur', 'YearDur', 'DateDur', 'Rate', 'TimeProb', 'RateProb', 'Time', 'years', 'months', 'weeks', 'days', 'perday', 'perweek', 'permonth', 'peryear']

#%% Base classes

class Date(pd.Timestamp):
    """
    Define a point in time, based on ``pd.Timestamp``

    Args:
        date (int/float/str/datetime): Any type of date input (ints and floats will be interpreted as years)
        kwargs (dict): passed to pd.Timestamp()

    **Examples**::

        ss.Date(2020) # Returns <2020-01-01>
        ss.Date(year=2020) # Returns <2020-01-01>
        ss.Date(year=2024.75) # Returns <2024-10-01>
        ss.Date('2024-04-04') # Returns <2024-04-04>
        ss.Date(year=2024, month=4, day=4) # Returns <2024-04-04>
    """

    def __new__(cls, *args, **kwargs):
        """ Check if a year was supplied, and preprocess it; complex due to pd.Timestamp implementation """
        single_year_arg = False
        if len(args) == 1:
            arg = args[0]
            if arg is None:
                return pd.Timestamp(None)
            elif isinstance(arg, pd.Timestamp):
                return cls._reset_class(sc.dcp(arg))
            elif sc.isnumber(arg):
                single_year_arg = True
        year_kwarg = len(args) == 0 and len(kwargs) == 1 and 'year' in kwargs
        if single_year_arg:
            return cls.from_year(args[0])
        if year_kwarg:
            return cls.from_year(kwargs['year'])

        # Otherwise, proceed as normal
        out = super(Date, cls).__new__(cls, *args, **kwargs)
        out = cls._reset_class(out)
        return out

    @classmethod
    def _reset_class(cls, obj):
        """ Manually reset the class from pd.Timestamp to ss.Date """
        obj.__class__ = Date
        return obj

    def __reduce__(self):
        # This function is internally used when pickling (rather than getstate/setstate)
        # due to Pandas implementing C functions for this. We can wrap the Pandas unpickler
        # with a function that wraps calls _reset_class to support unpickling Date
        # objects at the instance level
        unpickling_func, args = super().__reduce__()
        return (self.__class__._rebuild, (self.__class__, unpickling_func, args))

    @staticmethod
    def _rebuild(cls, unpickling_func, args):
        out = unpickling_func(*args)
        out = cls._reset_class(out)
        return out

    def __repr__(self, bracket=True):
        """ Show the date in brackets, e.g. <2024.04.04> """
        _ = ss.options.date_sep
        y = f'{self.year:04d}'
        m = f'{self.month:02d}'
        d = f'{self.day:02d}'
        string = y + _ + m + _ + d

        if self._time_repr != '00:00:00':
            string += ' ' + self._time_repr[:8]

        if bracket:
            string = '<' + string + '>'
        return string

    def __str__(self):
        """ Like repr, but just the date, e.g. 2024.04.04 """
        return self.__repr__(bracket=False)

    def replace(self, *args, **kwargs):
        """ Returns a new ss.Date(); pd.Timestamp is immutable """
        out = super().replace(*args, **kwargs)
        out = self.__class__._reset_class(out)
        return out

    def disp(self, **kwargs):
        """ Show the full object """
        return sc.pr(self, **kwargs)

    @classmethod
    def from_year(cls, year):
        """
        Convert an int or float year to a date.

        **Examples**::

            ss.Date.from_year(2020) # Returns <2020-01-01>
            ss.Date.from_year(2024.75) # Returns <2024-10-01>
        """

        if isinstance(year, int):
            return cls(year=year, month=1, day=1)
        else:
            year_start = pd.Timestamp(year=int(year), month=1, day=1)
            year_end = pd.Timestamp(year=int(year) + 1, month=1, day=1)
            return cls(year_start + year % 1 * (year_end - year_start))

    def to_year(self):
        """
        Convert a date to a floating-point year

        **Examples**::

            ss.Date('2020-01-01').to_year() # Returns 2020.0
            ss.Date('2024-10-01').to_year() # Returns 2024.7486
        """
        year_start = pd.Timestamp(year=self.year,month=1,day=1).timestamp()
        year_end = pd.Timestamp(year=self.year+1,month=1,day=1).timestamp()
        return self.year + (self.timestamp()-year_start)/(year_end-year_start)

    def __float__(self):
        return self.to_year()

    @property
    def years(self):
        # This matches Dur.years
        return self.to_year()

    def to_pandas(self):
        """ Convert to a standard pd.Timestamp instance """
        return pd.Timestamp(self.to_numpy()) # Need to convert to NumPy first or it doesn't do anything


    def __add__(self, other):
        if isinstance(other, np.ndarray):
            return np.vectorize(self.__add__)(other)

        if isinstance(other, DateDur):
            return Date(self.to_pandas() + other.period)
        elif isinstance(other, YearDur):
            return Date(self.to_year() + other.period)
        elif isinstance(other, pd.DateOffset):
            return Date(self.to_pandas() + other)
        elif isinstance(other, pd.Timestamp):
            raise TypeError('Cannot add a date to another date')
        elif sc.isnumber(other):
            raise TypeError(f'Attempted to add a number ({other}) to a Date, which is not supported. Only durations can be added to dates e.g., "ss.years({other})" or "ss.days({other})"')
        else:
            raise TypeError(f'Attempted to add an instance of {other.__class__.__name__} to a Date, which is not supported. Only durations can be added to dates.')

    def __sub__(self, other):

        if isinstance(other, np.ndarray):
            return np.vectorize(self.__sub__)(other)

        if isinstance(other, DateDur):
            return Date(self.to_pandas() - other.period)
        elif isinstance(other, YearDur):
            return Date(self.to_year() - other.period)
        elif isinstance(other, pd.DateOffset):
            return Date(self.to_pandas() - other)
        elif isinstance(other, Date):
            return YearDur(self.years-other.years)
        else:
            raise TypeError('Unsupported type')
    #
    def __radd__(self, other): return self.__add__(other)
    # def __iadd__(self, other): return self.__add__(other) # I think pd.Timestamp is immutable so these shouldn't be implemented?
    def __rsub__(self, other): return self.__sub__(other) # TODO: check if this should be reversed
    # def __isub__(self, other): return self.__sub__(other)

    def __lt__(self, other):
        if sc.isnumber(other):
            return self.to_year() < other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__lt__)(other)
        else:
            return super().__lt__(other)

    def __gt__(self, other):
        if sc.isnumber(other):
            return self.to_year() > other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__gt__)(other)
        else:
            return super().__gt__(other)

    def __le__(self, other):
        if sc.isnumber(other):
            return self.to_year() <= other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__le__)(other)
        else:
            return super().__le__(other)

    def __ge__(self, other):
        if sc.isnumber(other):
            return self.to_year() >= other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__ge__)(other)
        else:
            return super().__ge__(other)

    def __eq__(self, other):
        if sc.isnumber(other):
            return self.to_year() == other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__eq__)(other)
        else:
            return super().__eq__(other)

    def __ne__(self, other):
        if sc.isnumber(other):
            return self.to_year() != other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__ne__)(other)
        else:
            return super().__ne__(other)

    def __hash__(self):
        # As equality with float years is implemented, this allows interoperability with dicts that use float year keys
        return hash(self.to_year())

    # Convenience methods based on application usage

    @classmethod
    def from_array(cls, array):
        """
        Convert an array of float years into an array of Date instances

        :param array: An array of float years
        :return: An array of Date instances
        """
        return np.vectorize(cls)(array)

    @classmethod
    def arange(cls, low, high, step=1):
        """
        Construct an array of dates

        Functions similarly to np.arange, but returns Date objects

        Example usage:

        >>> Date.arange(2020, 2025)
            array([<2020.01.01>, <2021.01.01>, <2022.01.01>, <2023.01.01>,
                   <2024.01.01>], dtype=object)

        :param low: Lower bound - can be a date or a numerical year
        :param high: Upper bound - can be a date or a numerical year
        :param step: Assumes 1 calendar year steps by default
        :return: An array of Date instances
        """

        if isinstance(step, ss.DateDur):
            if not isinstance(low, Date):
                low = cls(low)
            if not isinstance(high, Date):
                high = cls(high)

            tvec = []
            t = low
            while t <= high:
                tvec.append(t)
                t += step
            return np.array(tvec)
        else:
            low = low.years if isinstance(low, Date) else low
            high = high.years if isinstance(high, Date) else high
            return cls.from_array(np.arange(low, high, step))


class Dur():
    # Base class for durations/periods
    # Subclasses for date-durations and fixed-durations

    # Conversion ratios from date-based durations to fixed durations
    ratios = sc.objdict(
        years=1,
        months=12,
        weeks=365.25/12/7, # This calculation allows a year to consist of 365.25 days. Note that this means a
        days=7,
        hours=24,
        minutes=60,
        seconds=60,
        milliseconds=1000,
        microseconds=1000,
        nanoseconds=1000,
    )

    def __new__(cls, *args, **kwargs):
        # Return
        if cls is Dur:
            if args:
                if isinstance(args[0], (pd.DateOffset, DateDur)):
                    return super().__new__(DateDur)
                elif isinstance(args[0], YearDur):
                    return super().__new__(YearDur)
                else:
                    assert len(args) == 1
                    return super().__new__(YearDur)
            else:
                return super().__new__(DateDur)
        return super().__new__(cls)

    # NB. Durations are considered to be equal if their year-equivalent duration is equal
    # That would mean that Dur(years=1)==Dur(1) returns True - probably less confusing than having it return False?
    def __hash__(self):
        return hash(self.years)

    def __eq__(self, other):
        return self.years == other.years

    def __float__(self):
        return float(self.years)

    @property
    def years(self):
        # Return approximate conversion into years
        raise NotImplementedError

    @property
    def to_numpy(self):
        raise NotImplementedError

    @property
    def is_variable(self):
        """
        Returns True if the duration has variable length

        Some date-based durations like years and months correspond to a variable
        period of time. This attribute captures whether or not the quantity has
        variable length.
        """
        raise NotImplementedError

    def __radd__(self, other):
        return self.__add__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return -1*self

    def __lt__(self, other):
        try:
            return self.years < other.years
        except:
            return self.years < other

    def __gt__(self, other):
        try:
            return self.years > other.years
        except:
            return self.years > other

    def __le__(self, other):
        try:
            return self.years <= other.years
        except:
            return self.years <= other

    def __ge__(self, other):
        try:
            return self.years >= other.years
        except:
            return self.years >= other

    def __eq__(self, other):
        try:
            return self.years == other.years
        except:
            return self.years == other

    def __ne__(self, other):
        try:
            return self.years != other.years
        except:
            return self.years != other

    def __truediv__(self, other):
        raise NotImplementedError('Dur subclasses are required to implement this method')

    def __rtruediv__(self, other):
        # If a Dur is divided by a Dur then we will call __truediv__
        # If a float is divided by a Dur, then we should return a rate
        # If a rate is divided by a Dur, then we will call Rate.__truediv__
        # We also need divide the duration by the numerator when calculating the rate
        if self.period == 0:
            raise ZeroDivisionError('Cannot divide by a duration of zero') # TODO: consider Rate(Dur(np.inf))
        elif sc.isnumber(other) and other == 0:
            return Rate(0)
        else:
            return Rate(other, self)

    def str(self):
        # This method should return a readable representation of the duration e.g., 'day', '2 days', 'year'
        # Not using __str__ because we still want print(Dur) to show the repr
        raise NotImplementedError('Dur subclasses are required to implement this method')

    @classmethod
    def arange(cls, low, high, step):
        """
        Construct an array of Dur instances

        For this function, the low, high, and step must ALL be specified, and they must
        all be Dur instances. Mixing Dur types (YearDur and DateDur) is permitted.

        :param low: Starting point e.g., ss.Dur(0)
        :param high:
        :param step:
        :return:
        """

        assert isinstance(low, Dur), 'Low input must be an ss.Dur'
        assert isinstance(high, Dur), 'High input must be an ss.Dur'
        assert isinstance(step, Dur), 'Step input must be an ss.Dur'

        tvec = []
        t = low
        while t <= high:
            tvec.append(t)
            t += step
        return np.array(tvec)


class YearDur(Dur):
    # Year based duration e.g., if requiring 52 weeks per year

    def __init__(self, years=0):
        """
        Construct a year-based fixed duration

        Supported inputs are
            - A float number of years (so can use values like 1/52 for 1 week)
            - A YearDur instance (as a copy-constructor)
            - A DateDur instance (to convert from date-based to fixed representation)

        :param years: float, YearDur, DateDur
        """
        if isinstance(years, Dur):
            self.period = years.years
        else:
            self.period = years

    def to_numpy(self):
        return self.period

    @property
    def years(self):
        return self.period # Needs to return float so matplotlib can plot it correctly

    @property
    def is_variable(self):
        return False

    def __add__(self, other):
        if isinstance(other, Dur):
            return self.__class__(self.period + other.years)
        elif isinstance(other, Date):
            return Date.from_year(other.to_year() + self.period)
        else:
            return self.__class__(self.period + other)

    def __sub__(self, other):
        if isinstance(other, Dur):
            return self.__class__(max(self.period - other.years, 0))
        elif isinstance(other, Date):
            return Date.from_year(other.to_year() - self.period)
        else:
            return self.__class__(max(self.period - other, 0))

    def __mul__(self, other: float):
        if isinstance(other, np.ndarray):
            return other*self # This means np.arange(2)*ss.years(1) and ss.years(1)*np.arange(2) both give [ss.years(1), ss.years(2)] as the output
        elif isinstance(other, Rate):
            return NotImplemented # Delegate to Rate.__rmul__
        elif isinstance(other, Dur):
            raise Exception('Cannot multiply a duration by a duration')
        elif isinstance(other, Date):
            raise Exception('Cannot multiply a duration by a date')
        return self.__class__(self.period*other)

    def __truediv__(self, other):
        if isinstance(other, Dur):
            return self.years/other.years
        elif isinstance(other, Rate):
            raise Exception('Cannot divide a duration by a rate')
        else:
            return self.__class__(self.period / other)

    def __repr__(self):
        return f'<YearDur: {self.period} years>'

    def str(self):
        if self.period == 1:
            return 'year'
        else:
            return f'{self.period} years'

    def __abs__(self):
        return self.__class__(abs(self.period))


class DateDur(Dur):
    # Date based duration e.g., if requiring a week to be 7 calendar days later

    def __init__(self, *args, **kwargs):
        """
        Create a date-based duration

        Supported arguments are
            - Single positional argument
                - A float number of years
                - A pd.DateOffset
                - A Dur instance
            - Keyword arguments
                - Argument names that match time periods in `Dur.ratios` e.g., 'years', 'months'
                  supporting floating point values. If the value is not an integer, the values
                  will be cascaded to smaller units using `Dur.ratios`

        If no arguments are specified, a DateDur with zero duration will be produced

        :param args:
        :param kwargs:
        """
        if args:
            assert not kwargs
            assert len(args) == 1
            if isinstance(args[0], pd.DateOffset):
                self.period = self._round_duration(args[0])
            elif isinstance(args[0], DateDur):
                self.period = args[0].period # pd.DateOffset is immutable so this should be OK
            elif isinstance(args[0], YearDur):
                self.period = self._round_duration({'years': args[0].years})
            elif sc.isnumber(args[0]):
                self.period = self._round_duration({'years': args[0]})
            else:
                raise TypeError('Unsupported input')
        else:
            self.period = self._round_duration(kwargs)

    @classmethod
    def _as_array(cls, dateoffset: pd.DateOffset) -> np.ndarray:
        """
        Return array representation of a pd.DateOffset

        In the array representation, each element corresponds to one of the time units
        in self.ratios e.g., a 1 year, 2 week duration would be [1, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0]

        :param dateoffset: A pd.DateOffset object
        :return: A numpy array with as many elements as `DateDur.ratios`
        """
        return np.array([dateoffset.kwds.get(k, 0) for k in cls.ratios.keys()])

    @classmethod
    def _as_args(cls, x) -> dict:
        """
        Return a dictionary representation of a DateOffset or DateOffset array

        This function takes in either a pd.DateOffset, or the result of `_as_array`, and returns
        a dictionary with the same keys as `DateDur.ratios` and the values from the input. The
        output of this function can be passed to the `Dur()` constructor to create a new DateDur object.

        :param x: A pd.DateOffset or an array with the same number of elements as `DateDur.ratios`
        :return: Dictionary with keys from `DateDur.ratios` and values from the input
        """
        if isinstance(x, pd.DateOffset):
            x = cls._as_array(x)
        return {k: v for k, v in zip(cls.ratios.keys(), x)}

    @property
    def years(self) -> float:
        """
        Return approximate conversion into years

        This allows interoperability with YearDur objects (which would typically be expected to
        occur if module parameters have been entered with `DateDur` durations, but the simulation
        timestep is in `YearDur` units).

        The conversion is based on `DateDur.ratios` which defines the conversion from each time unit
        to the next

        :return: A float representing the duration in years
        """
        denominator = 1
        years = 0
        for k, v in self.ratios.items():
            denominator *= v
            years += self.period.kwds.get(k, 0)/denominator
        return years

    @property
    def is_variable(self):
        if self.period.kwds.get('years',0) or self.period.kwds.get('months',0):
            return True
        else:
            return False

    @classmethod
    def _round_duration(cls, vals) -> pd.DateOffset:
        """
        Round a dictionary of duration values by overflowing remainders

        The input can be
            - A numpy array of length `cls.ratios` containing values in key order
            - A pd.DateOffset instance
            - A dictionary with keys from `cls.ratios`

        The output will be a pd.DateOffset with integer values, where non-integer values
        have been handled by overflow using the ratios in `cls.ratios`. For example, 2.5 weeks
        would first become 2 weeks and 0.5*7 = 3.5 days, and then become 3 days + 0.5*24 = 12 hours.

        Negative values are supported - -1.5 weeks for example will become (-1w, -3d, -12h)

        :return: A pd.DateOffset
        """
        if isinstance(vals, np.ndarray):
            d = sc.objdict({k:v for k,v in zip(cls.ratios, vals)})
        elif isinstance(vals, pd.DateOffset):
            d = sc.objdict.fromkeys(cls.ratios.keys(),0)
            d.update(vals.kwds)
        elif isinstance(vals, dict):
            d = sc.objdict.fromkeys(cls.ratios.keys(),0)
            d.update(vals)
        else:
            raise TypeError()

        for i in range(len(cls.ratios)-1):
            if d[i] < 0:
                d[i], remainder = divmod(d[i], 1)
                if remainder:
                    d[i]+=1
                    remainder -=1
            else:
                d[i], remainder = divmod(d[i], 1)
            d[i] = int(d[i])
            d[i+1] += remainder * cls.ratios[i+1]
        d[-1] = int(d[-1])

        return pd.DateOffset(**d)


    def _scale(self, dateoffset: pd.DateOffset, scale: float) -> pd.DateOffset:
        """
        Scale a pd.DateOffset by a factor

        This function will automatically cascade remainders to finer units using `DateDur.ratios` so for
        example 2.5 weeks would first become 2 weeks and 0.5*7 = 3.5 days,
        and then become 3 days + 0.5*24 = 12 hours.

        :param dateoffset: A pd.DateOffset instance
        :param scale: A float scaling factor (must be positive)
        :return: A pd.DateOffset instance scaled by the requested amount
        """
        return self._round_duration(self._as_array(dateoffset) * scale)

    def to_numpy(self):
        # TODO: This is a quick fix to get plotting to work, but this should do something more sensible
        return self.years
        # return self.period.to_numpy()

    def __pow__(self, other):
        raise Exception('Cannot multiply a duration by a duration')

    def __mul__(self, other: float):
        if isinstance(other, np.ndarray):
            return other*self
        elif isinstance(other, Rate):
            return NotImplemented # Delegate to Rate.__rmul__
        elif isinstance(other, Dur):
            raise Exception('Cannot multiply a duration by a duration')
        return self.__class__(self._scale(self.period, other))

    def __truediv__(self, other):
        if isinstance(other, DateDur):
            # We need to convert both dates into a common timebase to do the division
            # We *can* just do self.years/other.years, however this can cause numerical precision losses
            # yet at the same time is not necessary, because the common timebase depends on the durations.
            # For example, DateDur(weeks=1)/DateDur(days=1) should return 7, but we get 6.9999999 if we convert
            # both to years. Instead though, we can just convert days into weeks instead of days into years, and
            # then divide 1 week by (1/7) days.
            self_array = self._as_array(self.period)
            other_array = other._as_array(other.period)
            unit = np.argmax((self_array != 0) | (other_array != 0))
            denominator = 1
            a = 0
            b = 0
            for i, v in enumerate(self.ratios.values()):
                if i < unit:
                    continue
                if i > unit:
                    denominator *= v
                a += self_array[i] / denominator
                b += other_array[i] / denominator
            return a/b
        elif isinstance(other, Dur):
            return self.years/other.years
        elif isinstance(other, Rate):
            raise Exception('Cannot divide a duration by a rate')
        return self.__class__(self._scale(self.period, 1/other))

    def __repr__(self):
        if self.years == 0:
            return '<DateDur: 0>'
        else:
            labels = self.ratios.keys()
            vals = self._as_array(self.period).astype(float)

            time_portion = vals[4:]
            time_portion[-4] += time_portion[-1]*1e-9 + time_portion[-2]*1e-6 + time_portion[-3]*1e-3 # Collapse it down to seconds
            time_portion[-4] = int(vals[-4]*100)/100
            time_str = ':'.join(f'{np.round(v,1):02g}' for v in time_portion[:3])

            return '<DateDur: ' +  ','.join([f'{k}={int(v)}' for k, v in zip(labels[:4], vals[:4]) if v!=0]) + (f', +{time_str}' if time_str != '00:00:00' else '') + '>'

    def str(self):
        # Friendly representation e.g., 'day', '1 year, 2 months, 1 day'
        vals = self._as_array(self.period)

        # If we have only one nonzero value, return 'year', 'month', etc.
        if np.count_nonzero(vals == 1) == 1:
            for unit, val in zip(self.ratios, vals):
                if val == 1:
                    return unit[:-1]

        strs = [f'{v} {k[:-1] if abs(v) == 1 else k}' for k, v in zip(self.ratios, vals) if v != 0]
        return ', '.join(strs)


    def __add__(self, other):
        if isinstance(other, Date):
            return other + self.period
        elif isinstance(other, DateDur):
            return self.__class__(**self._as_args(self._as_array(self.period) + self._as_array(other.period)))
        elif isinstance(other, pd.DateOffset):
            return self.__class__(**self._as_args(self._as_array(self.period) + self._as_array(other)))
        elif isinstance(other, YearDur):
            kwargs = {k: v for k, v in zip(self.ratios, self._as_array(self.period))}
            kwargs['years'] += other.years
            return self.__class__(**kwargs)
        else:
            raise TypeError('For a DateDur instance, it is only possible to add or subtract dates, Dur objects, or pd.DateOffset objects')

    def __sub__(self, other):
        return self.__add__(-1*other)

    def __abs__(self):
        # Cannot implement this because it's ambiguous how to resolve cases like
        # DateDur(months=-1,days=1) - this is a sensible DateDur interpreted as 'go back 1 month, then go forward 1 day'
        # but just taking the absolute value of all of the components wouldn't work because this would be on average 1 month + 1 day
        # whereas it should be 1 month - 1 day. This could probably be resolved? But is an edge case, unlikely to be needed
        # (whereas abs(YearDur) arises when comparing dates, which automatically return YearDur)
        raise NotImplementedError('The absolute value of a DateDur instance is undefined as components (e.g., months, days) may have different signs.')

class Rate():
    """
    Store a value per unit time e.g., 2 per day
    - self.value - the numerator (e.g., 2) - a scalar float
    - self.period - the denominator (e.g., 1 day) - a Dur object
    """

    def __init__(self, value:float, period:Dur=None):
        if period is None:
            self.period = YearDur(1)
        elif isinstance(period, Dur):
            self.period = period
        else:
            self.period = Dur(period)

        assert sc.isnumber(value), 'Value must be a scalar number'
        self.value = value

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self.value} per {self.period.str()}>' # Use str to get the friendly representation

    def __mul__(self, other):
        if isinstance(other, np.ndarray):
            return other*self
        elif isinstance(other, Dur):
            return self.value*other/self.period
        else:
            return self.__class__(self.value*other, self.period)

    def __neg__(self):
        return -1*self

    def __add__(self, other):
        if self.__class__ == other.__class__:
            return self.__class__(self.value+other*self.period, self.period)
        elif not isinstance(other, Rate):
            if sc.isnumber(other) or isinstance(other, np.ndarray):
                raise TypeError('Only rates can be added to rates. This error most commonly occurs if the rate needs to be multiplied by `self.t.dt` to get a number of events per timestep.')
            else:
                raise TypeError('Only rates can be added to rates')
        else:
            raise TypeError('Can only add rates of identical types (e.g., Rate+Rate, TimeProb+TimeProb)')

    def __radd__(self, other): return self.__add__(other)

    def __sub__(self, other):
        if self.__class__ == other.__class__:
            return self.__class__(self.value-other*self.period, self.period)
        elif not isinstance(other, Rate):
            if sc.isnumber(other) or isinstance(other, np.ndarray):
                raise TypeError('Only rates can be added to rates. This error most commonly occurs if the rate needs to be multiplied by `self.t.dt` to get a number of events per timestep.')
            else:
                raise TypeError('Only rates can be added to rates')
        else:
            raise TypeError('Can only subtract rates of identical types (e.g., Rate+Rate, TimeProb+TimeProb)')

    def __eq__(self, other):
        return self.value == other.value/other.period*self.period

    def __rmul__(self, other): return self.__mul__(other)

    def __truediv__(self, other):
        # This is for <rate>/<other>
        if isinstance(other, Rate):
            # Convert the other rate onto our period, then divide the value
            return self.value/(other.value/other.period*self.period)
        elif isinstance(other, Dur):
            raise Exception('Cannot divide a rate by a duration')
        else:
            return self.__class__(self.value/other, self.period)

    def __rtruediv__(self, other):
        # If a float is divided by a rate
        return other * (self.period/self.value)

    def to_prob(self, dur, v=1):
        """
        Calculate a time-specific probability value

        This function is mainly useful for subclasses where the multiplication by a duration is non-linear
        (e.g., ``TimeProb``) and therefore it is important to apply the factor prior to multiplication by duration.
        This function avoids creating an intermediate array of rates, and is therefore much higher performance.

        e.g.

        >>> p = ss.TimeProb(0.05)*self.cd4*self.t.dt

        and

        >>> p = ss.TimeProb(0.05).scale(self.cd4,self.t.dt)

        are equivalent, except that the second one is (much) faster.

        :param v: The factor to scale the rate by. This factor is applied before multiplication by the duration
        :param dur: An ss.Dur instance to scale the rate by (often this is ``dt``)
        :return: A numpy float array of values
        """
        if isinstance(dur, np.ndarray):
            factor = (dur/self.period).astype(float)
        else:
            factor = dur/self.period
        return (self.value*v)*factor

class TimeProb(Rate):
    """
    ``TimeProb`` represents the probability of an event occurring during a
    specified period of time.

    The class is designed to allow conversion of a probability from one 
    duration to another through multiplication. However, the behavior of this
    conversion depends on the data type of the the object being multiplied.

    When multiplied by a duration (type ss.Dur), the underlying constant rate is
    calculated as
        ``rate = -np.log(1 - self.value)``.
    Then, the probability over the new duration is
        ``p = 1 - np.exp(-rate/factor)``,
    where ``factor`` is the ratio of the new duration to the original duration.

    For example,
    >>> p = ss.TimeProb(0.8, ss.years(1))
    indicates a 80% chance of an event occurring in one year.

    >>> p*ss.years(1)
    When multiplied by the original denominator, 1 year in this case, the
    probability remains unchanged, 80%.

    >>> p * ss.years(2)
    Multiplying ``p`` by ``ss.years(2)`` does not simply double the
    probability to 160% (which is not possible), but rather returns a new
    probability of 96% representing the chance of the event occurring at least
    once over the new duration of two years.

    However, the behavior is different when a ``TimeProb`` object is multiplied
    by a scalar or array. In this case, the probability is simply scaled. This scaling
    may result in a value greater than 1, which is not valid. For example,
    >>> p * 2
    raises an AssertionError because the resulting probability (160%) exceeds 100%.

    Use ``RateProb`` instead if ``TimeProb`` if you would prefer to directly
    specify the instantaneous rate.
    """

    def __init__(self, value, period=None):
        assert 0 <= value <= 1, 'Value must be between 0 and 1'
        return super().__init__(value, period)

    def __mul__(self, other):
        if isinstance(other, np.ndarray):
            return other*self # This means np.arange(2)*ss.years(1) and ss.years(1)*np.arange(2) both give [ss.years(1), ss.years(2)] as the output
        elif isinstance(other, Dur):
            if self.value == 0:
                return 0
            elif self.value == 1:
                return 1
            else:
                factor = self.period/other
                if factor == 1:
                    return self.value # Avoid expensive calculation and precision issues
                rate = -np.log(1 - self.value)
                return 1 - np.exp(-rate/factor)
        else:
            return self.__class__(self.value*other, self.period)

    def to_prob(self, dur, v=1):
        if isinstance(dur, np.ndarray):
            factor = (dur/self.period).astype(float)
        else:
            factor = dur/self.period
        rate = -np.log(1-(self.value*v))
        return 1-np.exp(-rate*factor)

    def __truediv__(self, other): raise NotImplementedError()
    def __rtruediv__(self, other): raise NotImplementedError()

class RateProb(Rate):
    """
    A ``RateProb`` represents an instantaneous rate of an event occurring. Rates
    must be non-negative, but need not be less than 1.

    Through multiplication, rate can be modified or converted to a probability,
    depending on the data type of the object being multiplied.

    When a ``RateProb`` is multiplied by a scalar or array, the rate is simply
    scaled. Such multiplication occurs frequently in epidemiological models,
    where the base rate is multiplied by "rate ratio" or "relative rate" to
    represent agents experiencing higher (multiplier > 1) or lower (multiplier <
    1) event rates.

    Alternatively, when a ``RateProb`` is multiplied by a duration (type
    ss.Dur), a probability is calculated. The conversion from rate to
    probability on multiplication by a duration is
        ``1 - np.exp(-rate/factor)``,
    where ``factor`` is the ratio of the multiplied duration to the original
    period (denominator).

    For example, consider
    >>> p = ss.RateProb(0.8, ss.years(1))
    When multiplied by a duration of 1 year, the calculated probability is
        ``1 - np.exp(-0.8)``, which is approximately 55%.
    >>> p*ss.years(1)

    When multiplied by a scalar, the rate is simply scaled.
    >>> p*2

    The difference between ``TimeProb`` and ``RateProb`` is subtle, but important. ``RateProb`` works directly
    with the instantaneous rate of an event occurring. In contrast, ``TimeProb`` starts with a probability and a duration,
    and the underlying rate is calculated. On multiplication by a duration,
    * RateProb: rate -> probability 
    * TimeProb: probability -> rate -> probability

    The behavior of both classes is depending on the data type of the object being multiplied.
    """
    def __init__(self, value, period=None):
        assert value >= 0, 'Value must be >= 0'
        return super().__init__(value, period)

    def __mul__(self, other):
        if isinstance(other, np.ndarray):
            return other*self
        elif isinstance(other, Dur):
            if self.value == 0:
                return 0
            else:
                factor = self.period/other
                return 1 - np.exp(-self.value/factor)
        else:
            return self.__class__(self.value*other, self.period)

    def to_prob(self, dur, v=1):
        if isinstance(dur, np.ndarray):
            factor = (dur/self.period).astype(float)
        else:
            factor = dur/self.period
        rate = self.value*v
        return 1-np.exp(-rate*factor)

    def __truediv__(self, other):
        if isinstance(other, Dur):
            raise NotImplementedError()
        return super().__truediv__(other)  # Fixed the call to super().__truediv__

    def __rtruediv__(self, other): raise NotImplementedError()

#%% Simulation time vectors

class Time(sc.prettyobj):
    """
    Handle time vectors for both simulations and modules.

    Each module can have its own time instance, in the case where the time vector
    is defined by absolute dates, these time vectors are by definition aligned. Otherwise
    they can be specified using Dur objects which express relative times (they can be added
    to a Date to get an absolute time)

    Args:
        start : ss.Date or ss.Dur
        stop : ss.Date if start is an ss.Date, or an ss.Dur if start is an ss.Dur
        dt (ss.Dur): Simulation step size
        pars (dict): if provided, populate parameter values from this dictionary
        parent (obj): if provided, populate missing parameter values from a 'parent" ``Time`` instance
        name (str): if provided, name the ``Time`` object
        init (bool): whether or not to immediately initialize the Time object
        sim (bool/Sim): if True, initializes as a sim-specific ``Time`` instance; if a Sim instance, initialize the absolute time vector

    The ``Time`` object, after initialization, has the following attributes:

    - ``ti`` (int): the current timestep
    - ``npts`` (int): the number of timesteps
    - ``tvec`` (array): time either as absolute `ss.Date` instances, or relative `ss.Dur` instances
    - ``yearvec`` (array): time represented as floating-point years

    **Examples**::

        t1 = ss.Time(start=2000, stop=2020, dt=1.0)
        t2 = ss.Time(start='2021-01-01', stop='2021-04-04', dt=ss.days(2))
    """

    # Allowable time arguments
    time_args = ['start', 'stop', 'dt']
    default_dur = Dur(50)
    default_start = Date(2000)
    default_dt = Dur(1)

    def __init__(self, start=None, stop=None, dt=None, dur=None, name=None):

        self.name = name
        self.start = start
        self.stop = stop
        self.dt = dt
        self.dur = dur
        self.ti = 0 # The time index, e.g. 0, 1, 2
        self._tvec    = None # The time vector for this instance in Date or Dur format
        self._yearvec = None # Time vector as floating point years
        self.initialized = False # Call self.init(sim) to initialise the object

        return

    @property
    def tvec(self):
        if self._tvec is None:
            raise Exception('Time object has not yet been not initialized - call `init()` before using the object')
        else:
            return self._tvec

    @property
    def yearvec(self):
        if self._yearvec is None:
            raise Exception('Time object has not yet been not initialized - call `init()` before using the object')
        else:
            return self._yearvec

    def __repr__(self):
        if self.initialized:
            return f'<Time t={self.tvec[self.ti]}, ti={self.ti}, {self.start}-{self.stop} dt={self.dt}>'
        else:
            return f'<Time (uninitialized)>'

    @property
    def timevec(self):
        # Backwards-compatibility function for now - TODO: consider deprecating?
        return self.tvec

    @property
    def datevec(self):
        # Backwards-compatibility function for now - TODO: consider deprecating?
        return self.tvec

    @property
    def npts(self):
        try:
            return self.tvec.shape[0]
        except:
            return 0

    @property
    def dt_year(self):
        return self.dt.years

    def __bool__(self):
        """ Always truthy """
        return True

    def __len__(self):
        """ Length is the number of timepoints """
        return self.npts

    @property
    def ready(self):
        """ Check if all parameters are in place to be initialized """
        return not any([getattr(self, k) is None for k in self.time_args])

    @property
    def is_absolute(self):
        """
        Check whether the fundamental simulation unit is absolute

        A time vector is absolute if the start is a Date rather than a Dur
        A relative time vector can be made absolute by adding a Date to it
        """
        try:
            return isinstance(self.start, Date)
        except:
            return False

    def update(self, pars=None, parent=None, reset=True, force=None, **kwargs):
        """ Reconcile different ways of supplying inputs """
        pars = sc.mergedicts(pars)
        stale = False

        for key in self.time_args:
            current_val = getattr(self, key, None)
            parent_val = getattr(parent, key, None)
            kw_val = kwargs.get(key)
            par_val = pars.get(key)

            # Special handling for dt: don't inherit dt if the units are different
            if key == 'dt':
                if isinstance(parent, Time):
                    if parent.unit != self.unit:
                        parent_val = 1.0

            if force is False: # Only update missing (None) values
                val = sc.ifelse(current_val, kw_val, par_val, parent_val)
            elif force is None: # Prioritize current value
                val = sc.ifelse(kw_val, par_val, current_val, parent_val)
            elif force is True: # Prioritize parent value
                val = sc.ifelse(kw_val, par_val, parent_val, current_val)
            else:
                errormsg = f'Invalid value {force} for force: must be False, None, or True'
                raise ValueError(errormsg)

            if val != current_val:
                setattr(self, key, val)
                stale = True

        if stale and reset and self.initialized:
            self.init()
        return

    def init(self, sim=None):
        """ Initialize all vectors """

        if sim is not None:
            self.dt = sc.ifelse(self.dt, sim.t.dt, sim.pars.dt)
            self.start = sc.ifelse(self.start, sim.t.start, sim.pars.start)
            self.stop = sc.ifelse(self.stop, sim.t.stop, sim.pars.stop)
            self.dur = sc.ifelse(self.dur, sim.t.dur, sim.pars.dur)

        if sc.isnumber(self.dur):
            self.dur = Dur(self.dur)

        match (self.start, self.stop, self.dur):
            case (None, None, None):
                start = self.default_start
                dur = self.default_dur
                stop = start+dur

            case (start, None, None):
                start = Date(start)
                dur = self.default_dur
                stop = start+dur

            case (None, stop, None):
                if isinstance(stop, Dur):
                    start = stop.__class__(0)
                elif sc.isnumber(stop) and stop < 1900:
                    stop = Dur(stop)
                    start = Dur(0)
                else:
                    stop = Date(stop)
                    start = self.default_start
                dur = stop-start

            case (None, None, dur):
                start = self.default_start
                stop = start+dur

            case (start, None, dur):
                start = Date(start)
                stop = start+dur

            case (None, stop, dur):
                if sc.isnumber(stop) and stop < 1900:
                    stop = Dur(stop)
                else:
                    stop = Date(stop)

            case (start, stop, dur):
                # Note that this block will run if dur is None and if it is not None, which is fine because
                # we are ignoring dur in this case (if the user specifies start and stop, they'll be used)
                if sc.isstring(start):
                    start = Date(start)
                if sc.isstring(stop):
                    stop = Date(stop)

                if sc.isnumber(start) and sc.isnumber(stop):
                    if stop < 1900:
                        start = Dur(start)
                        stop = Dur(stop)
                    else:
                        start = Date(start)
                        stop = Date(stop)
                elif sc.isnumber(start):
                    start = stop.__class__(start)
                elif sc.isnumber(stop):
                    stop = start.__class__(stop)
                dur = stop-start
            case _:
                raise Exception('Failed to match start, stop, and dur') # This should not occur

        assert isinstance(start, (Date, Dur)), 'Start must be a Date or Dur'
        assert isinstance(stop, (Date, Dur)), 'Stop must be a Date or Dur'
        assert type(start) is type(stop), 'Start and stop must be the same type'
        assert start <= stop, 'Start must be before stop'

        self.start = start
        self.stop = stop
        self.dur = dur

        if self.dt is None:
            self.dt = self.default_dt

        if sc.isnumber(self.dt):
            self.dt = Dur(self.dt)

        if sim is not None:
            if self.is_absolute != sim.t.is_absolute:
                raise Exception('Cannot mix absolute/relative times across modules')

        # We need to populate both the tvec (using dates) and the yearvec (using years). However, we
        # need to decide which of these quantities to prioritise considering that the calendar dates
        # don't convert consistently into fractional years due to varying month/year lengths. We will
        # prioritise one or the other depending on what type of quantity the user has specified for dt
        if isinstance(self.dt, YearDur):
            # If dt has been specified as a YearDur then preference setting fractional years. So first
            # calculate the fractional years, and then convert them to the equivalent dates
            self._yearvec = np.round(self.start.years + np.arange(0, self.stop.years - self.start.years + self.dt.years, self.dt.years), 12)  # Subtracting off self.start.years in np.arange increases floating point precision for that part of the operation, reducing the impact of rounding
            if isinstance(self.stop, Dur):
                self._tvec = np.array([self.stop.__class__(x) for x in self._yearvec])
            else:
                self._tvec = np.array([Date(x) for x in self._yearvec])
        else:
            # If dt has been specified as a DateDur then preference setting dates. So first
            # calculate the dates/durations, and then convert them to the equivalent fractional years
            if isinstance(self.stop, Dur):
                self._tvec = ss.Dur.arange(self.start, self.stop, self.dt)
            else:
                self._tvec = ss.Date.arange(self.start, self.stop, self.dt)
            self._yearvec = np.array([x.years for x in self._tvec])

        self.initialized = True
        return self

    def now(self, key=None):
        """
        Get the current simulation time

        Args:
            which (str): which type of time to get: default (None), "year", "date", "tvec", or "str"

        **Examples**::

            t = ss.Time(start='2021-01-01', stop='2022-02-02', dt=1, unit='week')
            t.ti = 25
            t.now() # Returns <2021-06-25>
            t.now('date') # Returns <2021-06-25>
            t.now('year') # Returns 2021.479
            t.now('str') # Returns '2021-06-25'
        """
        if key in [None, 'none', 'time', 'str']:
            vec = self.tvec
        elif key == 'year':
            vec = self.yearvec
        else:
            errormsg = f'Invalid key "{key}": must be None, abs, date, or year'
            raise ValueError(errormsg)

        if 0 <= self.ti < len(vec):
            now = vec[self.ti]
        else:
            now = self.tvec[0] + self.dt*self.ti
            if key == 'year':
                now = float(now)

        if key == 'str':
            now = str(now)

        return now


#%% Convenience functions

def years(x: float) -> Dur:
    return Dur(x)

def months(x: float) -> Dur:
    return Dur(months=x)

def weeks(x: float) -> Dur:
    return Dur(weeks=x)

def days(x: float) -> Dur:
    return Dur(days=x)

def perday(v):
    """Shortcut to specify rate per calendar day"""
    return Rate(v, Dur(days=1))

def perweek(v):
    """Shortcut to specify rate per calendar week"""
    return Rate(v, Dur(weeks=1))

def permonth(v):
    """Shortcut to specify rate per calendar month"""
    return Rate(v, Dur(months=1))

def peryear(v):
    """Shortcut to specify rate per numeric year"""
    return Rate(v, YearDur(1))


#%% Demos/tests

if __name__ == '__main__':

    from starsim.time import *   # Import the classes from Starsim so that Dur is an ss.Dur rather than just a bare Dur etc.
    import starsim as ss

    def loc(module, sim, uids):
        return np.array([Dur(x) for x in range(uids)])


    module = sc.objdict(t=sc.objdict(dt=Dur(days=1)))
    d = ss.normal(loc, Dur(days=1), module=module, strict=False)
    d.init()
    d.rvs(10)

    Date(1500)
    Date(1500.1)
    YearDur(1)*np.arange(5)
    np.arange(5)*YearDur(1)


    t = Time(start=2001, stop=2003, dt=ss.years(0.1))

    Dur(weeks=1)/Dur(days=1)

    ss.perweek(1)+ss.perday(1)


    ss.peryear(10)
    # module = sc.objdict(t=sc.objdict(dt=Dur(weeks=1)))
    # d = ss.poisson(ss.peryear(10), module=module, strict=False)
    # d.init()
    # d.rvs(5)


    DateDur(weeks=1) - DateDur(days=1)

    Date(2020.1)

    ss.Date(2050) - ss.Date(2020)


    print(Dur(weeks=1)+Dur(1/52))
    print(Dur(1/52)+Dur(weeks=1))
    print(Dur(weeks=1)/Dur(1/52))
    print(DateDur(YearDur(2/52)))

    # Date('2020-01-01') + Dur(weeks=52)  # Should give us 30th December 2020
    Date('2020-01-01') + 2.5*Dur(weeks=1)  # Should give us 30th December 2020
    Date('2020-01-01') + 52*Dur(weeks=1)  # Should give us 30th December 2020
    Date('2020-01-01') + 52*Dur(1/52) # Should give us 1st Jan 2021


    import pickle
    x = Date('2020-01-01')
    s = pickle.dumps(x)
    pickle.loads(s)

    t = Time(Date('2020-01-01'), Date('2020-06-01'), Dur(days=1))
    t = Time(Date('2020-01-01'), Date('2020-06-01'), Dur(months=1)) # Allowed

    t = Time(Dur(days=0), Dur(days=30), Dur(days=1))
    t = Time(Dur(days=0), Dur(months=1), Dur(days=30))
    t = Time(Dur(days=0), Dur(years=1), Dur(weeks=1))
    t = Time(Dur(days=0), Dur(years=1), Dur(months=1)) # Not allowed
    t = Time(Dur(days=0), Dur(years=1), Dur(months=1)) # Not allowed


    t = Time(Dur(0), Dur(1), Dur(1/12)).init()
    t.tvec+Date(2020)

    # Date('2020-01-01')+50*Dur(days=1)
    t = Time(Date('2020-01-01'), Date('2030-06-01'), Dur(days=1))

    t = Time(Date(2020), Date(2030.5), Dur(0.1)).init()
    t.tvec + YearDur(1)

    print(1/YearDur(1))
    print(2/YearDur(1))
    print(4/YearDur(1))
    print(4/DateDur(1))
    print(0.5/DateDur(1))

    print(perday(5)*Dur(days=1))


    # time_prob
    p = TimeProb(0.1, Dur(years=1))
    p*Dur(years=2)
    p * Dur(0.5)
    p * Dur(months=1)

    p = TimeProb(0.1, Dur(1))
    p*Dur(years=2)
    p * Dur(0.5)
    p * Dur(months=1)

    2/Rate(0.25)
    1/(2*Rate(0.25))
    Rate(0.5)/Rate(1)


    # Dists
    module = sc.objdict(t=sc.objdict(dt=Dur(days=1)))
    d = ss.normal(Dur(days=6), Dur(days=1), module=module, strict=False)
    d.init()
    d.rvs(5)

    module = sc.objdict(t=sc.objdict(dt=Dur(weeks=1)))
    d = ss.normal(Dur(days=6), Dur(days=1), module=module, strict=False)
    d.init()
    d.rvs(5)

    Rate(4)/Rate(3)

    2*Rate(5)


    ss.perday(1)+ss.perweek(1)

    ss.perweek(1)+ss.perday(1)


    # Another example
    # 5/DateDur(weeks=1)
    # Out[3]: <Rate: per days=1,hours=9,minutes=36> (approx. 259.99999999999994 per year)
    # 259/365
    # Out[4]: 0.7095890410958904
    # 5/7*365
    # Out[5]: 260.7142857142857
    # r = 5/DateDur(weeks=1)
    # r*DateDur(days=1)
    # Out[7]: 0.7142857142857142
    # r*DateDur(days=5)
    # Out[8]: 3.5714285714285707
    # r*DateDur(days=7)
    # Out[9]: 4.999999999999999
    # r*DateDur(weeks=1)



"""
Simulation and module timelines
"""
import sciris as sc
import numpy as np
import starsim as ss

__all__ = ['Timeline']


class Timeline:
    """
    Handle time vectors and sequencing ("timelines") for both simulations and modules.

    Each module can have its own time instance, in the case where the time vector
    is defined by absolute dates, these time vectors are by definition aligned. Otherwise
    they can be specified using Dur objects which express relative times (they can be added
    to a date to get an absolute time)

    Args:
        start (str/int/float/ss.date/ss.Dur): when the simulation/module starts, e.g. '2000', '2000-01-01', 2000, ss.date(2000), or ss.years(2000)
        stop (str/int/float/ss.date/ss.Dur): when the simulation/module ends (note: if start is a date, stop must be too)
        dt (int/float/ss.Dur): Simulation step size
        name (str): if provided, name the `Timeline` object
        init (bool): whether or not to immediately initialize the `Timeline` object (by default, yes if start and stop or start and dur are provided; otherwise no)
        sim (Sim): if provided, initialize the `Timeline` with this as the parent (i.e. populating missing values)

    The `Timeline` object, after initialization, has the following attributes:

    - `ti` (int): the current timestep
    - `npts` (int): the number of timesteps
    - `tvec` (array): time either as absolute `ss.date` instances, or relative `ss.Dur` instances
    - `yearvec` (array): time represented as floating-point years

    **Examples**:

        t1 = ss.Timeline(start=2000, stop=2020, dt=1.0)
        t2 = ss.Timeline(start='2021-01-01', stop='2021-04-04', dt=ss.days(2))
    """

    # Allowable time arguments
    time_args = ['start', 'stop', 'dt']

    def __init__(self, start=None, stop=None, dt=None, dur=None, name=None, init=None, sim=None):
        # Store inputs
        self.name = name
        self.start = start
        self.stop = stop
        self.dt = dt
        self.dur = dur

        # Set defaults
        default_type = ss.years # The default Dur type, e.g. ss.years; this is reset after all the parameters are reconciled
        self.default_type  = default_type
        self.default_dur   = default_type(50)
        self.default_start = default_type(2000)
        self.default_dt    = default_type(1.0)
        self.calendar_year_threshold = 1900.0 # Value at which to switch over from interpreting a value from a year duration to a calendar year

        # Populated later
        self.ti = 0 # The time index, e.g. 0, 1, 2
        self.tvec    = None # The time vector for this instance in date or Dur format
        self.yearvec = None # Time vector as floating point years
        self.datevec = None # The time vector as date objects
        self.initialized = False # Call self.init(sim) to initialize the object

        # Decide whether to initialized: we're asked, a sim is provided, or arguments are supplied directly
        if init or sim or init is None and sum([x is not None for x in [start, stop, dur]]) >= 2:
            self.init(sim)
        return

    def __repr__(self):
        if self.initialized:
            return f'<Timeline t={self.tvec[self.ti]}, ti={self.ti}, {self.start}-{self.stop} dt={self.dt}>'
        else:
            return '<Timeline (uninitialized)>'

    def disp(self):
        return sc.pr(self)

    @property
    def timevec(self):
        warnmsg = 'In Starsim v3.0, use t.tvec instead of t.timevec'
        ss.warn(warnmsg)
        return self.tvec

    @property
    def npts(self):
        try:
            return self.tvec.shape[0]
        except:
            return 0

    @property
    def dt_year(self):
        return self.dt.years

    def __bool__(self):
        """ Always truthy """
        return True

    def __len__(self):
        """ Length is the number of timepoints """
        return self.npts

    @property
    def ready(self):
        """ Check if all parameters are in place to be initialized """
        return not any([getattr(self, k) is None for k in self.time_args])

    @property
    def finished(self):
        """
        Check if the simulation is finished, i.e. we're at the last time point
        (note, this does not distinguish whether we are at the beginning or end
        of the last time point, so use with caution!)
        """
        return self.ti == self.npts-1

    @property
    def is_absolute(self):
        """
        Check whether the fundamental simulation unit is absolute

        A time vector is absolute if the start is a date rather than a Dur
        A relative time vector can be made absolute by adding a date to it.
        """
        try:
            return isinstance(self.start, ss.date)
        except:
            return False

    def now(self, key=None):
        """
        Get the current simulation time

        Args:
            key (str): which type of time to get: default (None), "year", "date", "tvec", or "str"

        **Examples**:

            t = ss.Timeline(start='2021-01-01', stop='2022-02-02', dt='week')
            t.ti = 25
            t.now() # Returns <2021-06-25>
            t.now('date') # Returns <2021-06-25>
            t.now('year') # Returns 2021.479
            t.now('str') # Returns '2021-06-25'
        """
        if key in [None, 'none', 'time', 'str']:
            vec = self.tvec
        elif key == 'year':
            vec = self.yearvec
        elif key == 'date':
            vec = self.datevec
        else:
            errormsg = f'Invalid key "{key}": must be None, date, or year'
            raise ValueError(errormsg)

        if 0 <= self.ti < len(vec):
            now = vec[self.ti]
        else:
            now = self.tvec[0] + self.dt*self.ti
            if key == 'year':
                now = float(now)

        if key == 'str':
            now = str(now)

        return now

    def update(self, pars=None, parent=None, reset=True, force=None, **kwargs):
        """ Reconcile different ways of supplying inputs """
        pars = sc.mergedicts(pars)
        stale = False

        for key in self.time_args:
            current_val = getattr(self, key, None)
            parent_val = getattr(parent, key, None)
            kw_val = kwargs.get(key)
            par_val = pars.get(key)

            # Special handling for dt: don't inherit dt if the units are different
            if key == 'dt':
                if isinstance(parent, Timeline):
                    if parent.unit != self.unit:
                        parent_val = 1.0

            if force is False: # Only update missing (None) values
                val = sc.ifelse(current_val, kw_val, par_val, parent_val)
            elif force is None: # Prioritize current value
                val = sc.ifelse(kw_val, par_val, current_val, parent_val)
            elif force is True: # Prioritize parent value
                val = sc.ifelse(kw_val, par_val, parent_val, current_val)
            else:
                errormsg = f'Invalid value {force} for force: must be False, None, or True'
                raise ValueError(errormsg)

            if val != current_val:
                setattr(self, key, val)
                stale = True

        if stale and reset and self.initialized:
            self.init()
        return

    def reconcile_args(self, sim=None):
        """ Reconcile the different options for the input parameters """

        def is_calendar_year(val):
            """ Whether a number should be interpreted as a calendar year -- by default, a number greater than 1900 """
            return sc.isnumber(val) and val > self.calendar_year_threshold

        if sim is not None:
            self.dt    = sc.ifelse(self.dt,    sim.t.dt,    sim.pars.dt)
            self.start = sc.ifelse(self.start, sim.t.start, sim.pars.start)
            self.stop  = sc.ifelse(self.stop,  sim.t.stop,  sim.pars.stop)
            self.dur   = sc.ifelse(self.dur,   sim.t.dur,   sim.pars.dur)

        # Check to see if any inputs were provided as durations: if so, reset the default type
        for arg in [self.start, self.stop, self.dur, self.dt]:
            if isinstance(arg, ss.Dur):
                self.default_type = type(arg)
                break # Stop at the first one

        # Ensure dur is valid
        if sc.isnumber(self.dur):
            self.dur = self.default_type(self.dur)
        assert self.dur is None or isinstance(self.dur, ss.Dur), 'Timeline.dur must be a number, a Dur object or None'

        # Now, figure out start, stop, and dur
        match (self.start, self.stop, self.dur):
            case (None, None, None):
                start = self.default_start
                dur = self.default_dur
                stop = start + dur

            case (start, None, None):
                if isinstance(start, ss.Dur):
                    pass  # Already a Dur which is fine
                elif is_calendar_year(start):
                    start = ss.date(start)
                else:
                    start = ss.Dur(start)
                dur = self.default_dur
                stop = start + dur

            case (None, stop, None):
                if isinstance(stop, ss.Dur):
                    start = stop.__class__(value=0)
                elif is_calendar_year(stop):
                    stop = ss.date(stop)
                    start = ss.date(self.default_start)
                else:
                    stop = ss.Dur(stop)
                    start = ss.Dur(0)
                dur = stop - start

            case (None, None, dur):
                start = self.default_start
                stop = start+dur

            case (start, None, dur):
                if isinstance(start, ss.Dur):
                    pass
                elif is_calendar_year(start):
                    start = ss.Dur(start)
                else:
                    start = ss.date(start)
                stop = start + dur

            case (None, stop, dur):
                if isinstance(stop, ss.Dur):
                    pass
                elif is_calendar_year(stop):
                    stop = ss.date(stop)
                else:
                    stop = ss.Dur(stop)
                start = stop - dur

            case (start, stop, dur):
                # Note that this block will run if dur is None and if it is not None, which is fine because
                # we are ignoring dur in this case (if the user specifies start and stop, they'll be used)
                if sc.isstring(start):
                    start = ss.date(start)
                if sc.isstring(stop):
                    stop = ss.date(stop)

                if sc.isnumber(start) and sc.isnumber(stop):
                    if is_calendar_year(start):
                        start = ss.date(start)
                        stop = ss.date(stop)
                    else:
                        start = ss.Dur(start)
                        stop = ss.Dur(stop)
                elif sc.isnumber(start):
                    start = stop.__class__(start)
                elif sc.isnumber(stop):
                    stop = start.__class__(stop)
                dur = stop - start
            case _:
                errormsg = f'Failed to match {start = }, {stop = }, and {dur = } to any known pattern. You can use numbers, strings, ss.date, or ss.Dur objects.'
                raise ValueError(errormsg) # This should not occur

        start_type = type(start)
        stop_type = type(start)
        assert isinstance(start, (ss.date, ss.Dur)), f'Start must be ss.date or ss.Dur, not {start_type}'
        assert isinstance(stop, (ss.date, ss.Dur)), f'Stop must be ss.date or ss.Dur, not {stop_type}'
        assert start_type is stop_type, f'Start and stop must be the same type, not {start_type} and {stop_type}'
        assert start <= stop, f'Start must be before stop, not {start} and {stop}'

        # Figure out type and store everything
        if issubclass(start_type, ss.Dur):
            self.default_type = start_type # Now that we've reconciled everything, reset the default type if needed
        self.start = start
        self.stop = stop
        self.dur = dur

        if self.dt is None:
            self.dt = self.default_dt

        if sc.isnumber(self.dt):
            self.dt = self.default_type(self.dt)
        return

    def init(self, sim=None):
        """ Initialize all vectors """

        # Handle start, stop, dt, dur
        self.reconcile_args(sim)

        if sim is not None:
            if self.is_absolute != sim.t.is_absolute:
                errormsg = 'Cannot mix absolute/relative times across modules and the sim: both must be either dates or e.g. year/day ranges'
                raise Exception(errormsg)

        # The sim is provided and matches the current object: copy from the sim
        if sim is not None and sim.t.initialized and \
                all([type(getattr(self, key)) == type(getattr(sim.t, key)) for key in ('start', 'stop', 'dt')]) and \
                all([getattr(self, key) == getattr(sim.t, key) for key in ('start', 'stop', 'dt')]):
            self.yearvec = sc.dcp(sim.t.yearvec)
            self.tvec    = sc.dcp(sim.t.tvec)
            self.datevec = sc.cp(sim.t.datevec) # ss.Dates are immutable so only need shallow copy
            self.initialized = True
            return self

        # We need to make the tvec (dates/Durs) the yearvec (years), and the datevec (dates). However, we
        # need to decide which of these quantities to prioritise considering that the calendar dates
        # don't convert consistently into fractional years due to varying month/year lengths. We will
        # prioritise one or the other depending on what type of quantity the user has specified for start
        if self.start.years >= 1: # Years below 1 are not allowed
            self.datevec = ss.date.arange(self.start, self.stop, self.dt)

        if isinstance(self.start, ss.Dur): # Use durations
            self.yearvec = np.round(self.start.years + np.arange(0, self.stop.years - self.start.years + self.dt.years, self.dt.years), 12)  # Subtracting off self.start.years in np.arange increases floating point precision for that part of the operation, reducing the impact of rounding
            self.tvec = np.array([self.default_type(value=x) for x in self.yearvec]) # TODO: refactor
        elif isinstance(self.start, ss.date):
            self.tvec = self.datevec
            self.yearvec = np.array([x.years for x in self.datevec])
        else:
            errormsg = f'Unexpected start {self.start}: expecting ss.Dur or ss.Date'
            raise TypeError(errormsg)

        if self.start.years < 1: # If we didn't initialize the datevec before, do so now
            self.datevec = self.tvec

        self.initialized = True
        return self



#!/usr/bin/env python3
"""
Larger test of sim performance.
"""

import sciris as sc
import starsim as ss
import networkx as nx
import numpy as np
import pandas as pd 


repeats = 10
n_agents = 1_000

# Vary the infection duration from 1 to 100 days
duration = [1,
    2,
    5,
    10,
    15,
    25,
    50,
    75,
    90,
    100
    ]

def make_run_sim(duration,r):
    """ Make and run a decent-sized default simulation """
    # Make the people
    ppl = ss.People(n_agents=n_agents)
    demographics = [
    ss.Births(pars={'birth_rate': 20}),
    ss.Deaths(pars={'death_rate': 15})
    ]

    sir = ss.SIR()
    ss.SIR(pars=dict( dur_inf=duration[r], beta=0.2, init_prev=0.4, p_death=0.2))
    networks=ss.DynamicNetwork()

    # Make the sim
    sim = ss.Sim(people=ppl, networks=networks, demographics=demographics, diseases=sir)

    # Run the sim
    sim.run()
    
    return sim


if __name__ == '__main__':
    
    edge_count=[]    
    T = sc.timer()
    cpr = sc.cprofile()
    cpr.start()
    for r in range(repeats):    
        make_run_sim(duration, r)
        T.tt(f'Trial {r+1}/{repeats}')      

    sc.heading(f'Average: {T.mean()*1000:0.0f} ± {T.std()/len(T)**0.5*1000:0.0f} ms')
    cpr.stop()
    df=cpr.to_df()
    df.to_csv('DynamicNet Test Profile.csv', index=False)
    print("Run "+str(repeats)+" Mem "+str(sc.memload()))
    print("RAM "+str(repeats)+" "+str(sc.checkram()))
